from __future__ import annotations

from math import cos, pi, sin
from typing import Sequence

from rich.console import Console, ConsoleOptions, RenderResult
from rich.segment import Segment
from rich.style import Style

from textual.color import Color, Gradient


class VerticalGradient:
    """Draw a vertical gradient."""

    def __init__(self, color1: str, color2: str) -> None:
        self._color1 = Color.parse(color1)
        self._color2 = Color.parse(color2)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        width = options.max_width
        height = options.height or options.max_height
        color1 = self._color1
        color2 = self._color2
        default_color = Color(0, 0, 0).rich_color
        from_color = Style.from_color
        blend = color1.blend
        rich_color1 = color1.rich_color
        for y in range(height):
            line_color = from_color(
                default_color,
                (
                    blend(color2, y / (height - 1)).rich_color
                    if height > 1
                    else rich_color1
                ),
            )
            yield Segment(f"{width * ' '}\n", line_color)


class LinearGradient:
    """Render a linear gradient with a rotation.

    Args:
        angle: Angle of rotation in degrees.
        stops: List of stop consisting of pairs of offset (between 0 and 1) and color.

    """

    def __init__(
        self, angle: float, stops: Sequence[tuple[float, Color | str]]
    ) -> None:
        self.angle = angle
        self._stops = [
            (stop, Color.parse(color) if isinstance(color, str) else color)
            for stop, color in stops
        ]
        self._color_gradient = Gradient(*self._stops)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        width = options.max_width
        height = options.height or options.max_height

        angle_radians = -self.angle * pi / 180.0
        sin_angle = sin(angle_radians)
        cos_angle = cos(angle_radians)

        center_x = width / 2
        center_y = height

        new_line = Segment.line()

        _Segment = Segment
        get_color = self._color_gradient.get_rich_color
        from_color = Style.from_color

        for line_y in range(height):
            point_y = float(line_y) * 2 - center_y
            point_x = 0 - center_x

            x1 = (center_x + (point_x * cos_angle - point_y * sin_angle)) / width
            x2 = (
                center_x + (point_x * cos_angle - (point_y + 1.0) * sin_angle)
            ) / width
            point_x = width - center_x
            end_x1 = (center_x + (point_x * cos_angle - point_y * sin_angle)) / width
            delta_x = (end_x1 - x1) / width

            if abs(delta_x) < 0.0001:
                # Special case for verticals
                yield _Segment(
                    "▀" * width,
                    from_color(
                        get_color(x1),
                        get_color(x2),
                    ),
                )

            else:
                yield from [
                    _Segment(
                        "▀",
                        from_color(
                            get_color(x1 + x * delta_x),
                            get_color(x2 + x * delta_x),
                        ),
                    )
                    for x in range(width)
                ]

            yield new_line


if __name__ == "__main__":
    from rich import print

    COLORS = [
        "#881177",
        "#aa3355",
        "#cc6666",
        "#ee9944",
        "#eedd00",
        "#99dd55",
        "#44dd88",
        "#22ccbb",
        "#00bbcc",
        "#0099cc",
        "#3366bb",
        "#663399",
    ]

    stops = [(i / (len(COLORS) - 1), Color.parse(c)) for i, c in enumerate(COLORS)]

    print(LinearGradient(25, stops))

    from time import time

    from textual.app import App, ComposeResult
    from textual.widgets import Static

    class GradientApp(App):
        CSS = """
        Screen {
            background: transparent;
            align: center middle;
        }

        Static {
            padding: 2 4;
            background: $panel;
            width: 50;
        }

        """

        def compose(self) -> ComposeResult:
            yield Static("Gradients are fast now :-) ")

        def render(self):
            return LinearGradient(time() * 90, stops)

        def on_mount(self) -> None:
            self.set_interval(1 / 30, self.refresh)

    app = GradientApp()
    app.run()



"""

This module contains the `Binding` class and related objects.

See [bindings](/guide/input#bindings) in the guide for details.
"""

from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from typing import TYPE_CHECKING, Iterable, Iterator, Mapping, NamedTuple

import rich.repr

from textual.keys import _character_to_key

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from textual.dom import DOMNode

BindingType: TypeAlias = "Binding | tuple[str, str] | tuple[str, str, str]"
"""The possible types of a binding found in the `BINDINGS` class variable."""

BindingIDString: TypeAlias = str
"""The ID of a Binding defined somewhere in the application.

Corresponds to the `id` parameter of the `Binding` class.
"""

KeyString: TypeAlias = str
"""A string that represents a key binding.

For example, "x", "ctrl+i", "ctrl+shift+a", "ctrl+j,space,x", etc.
"""

Keymap = Mapping[BindingIDString, KeyString]
"""A mapping of binding IDs to key strings, used for overriding default key bindings."""


class BindingError(Exception):
    """A binding related error."""


class NoBinding(Exception):
    """A binding was not found."""


class InvalidBinding(Exception):
    """Binding key is in an invalid format."""


@dataclass(frozen=True)
class Binding:
    """The configuration of a key binding."""

    key: str
    """Key to bind. This can also be a comma-separated list of keys to map multiple keys to a single action."""
    action: str
    """Action to bind to."""
    description: str = ""
    """Description of action."""
    show: bool = True
    """Show the action in Footer, or False to hide."""
    key_display: str | None = None
    """How the key should be shown in footer.

    If None, the display of the key will use the result of `App.get_key_display`.

    If overridden in a keymap then this value is ignored.
    """
    priority: bool = False
    """Enable priority binding for this key."""
    tooltip: str = ""
    """Optional tooltip to show in footer."""

    id: str | None = None
    """ID of the binding. Intended to be globally unique, but uniqueness is not enforced.

    If specified in the App's keymap then Textual will use this ID to lookup the binding,
    and substitute the `key` property of the Binding with the key specified in the keymap.
    """
    system: bool = False
    """Make this binding a system binding, which removes it from the key panel."""

    def parse_key(self) -> tuple[list[str], str]:
        """Parse a key into a list of modifiers, and the actual key.

        Returns:
            A tuple of (MODIFIER LIST, KEY).
        """
        *modifiers, key = self.key.split("+")
        return modifiers, key

    def with_key(self, key: str, key_display: str | None = None) -> Binding:
        """Return a new binding with the key and key_display set to the specified values.

        Args:
            key: The new key to set.
            key_display: The new key display to set.

        Returns:
            A new binding with the key set to the specified value.
        """
        return dataclasses.replace(self, key=key, key_display=key_display)

    @classmethod
    def make_bindings(cls, bindings: Iterable[BindingType]) -> Iterable[Binding]:
        """Convert a list of BindingType (the types that can be specified in BINDINGS)
        into an Iterable[Binding].

        Compound bindings like "j,down" will be expanded into 2 Binding instances.

        Args:
            bindings: An iterable of BindingType.

        Returns:
            An iterable of Binding.
        """
        bindings = list(bindings)
        for binding in bindings:
            # If it's a tuple of length 3, convert into a Binding first
            if isinstance(binding, tuple):
                if len(binding) not in (2, 3):
                    raise BindingError(
                        f"BINDINGS must contain a tuple of two or three strings, not {binding!r}"
                    )
                # `binding` is a tuple of 2 or 3 values at this point
                binding = Binding(*binding)  # type: ignore[reportArgumentType]

            # At this point we have a Binding instance, but the key may
            # be a list of keys, so now we unroll that single Binding
            # into a (potential) collection of Binding instances.
            for key in binding.key.split(","):
                key = key.strip()
                if not key:
                    raise InvalidBinding(
                        f"Can not bind empty string in {binding.key!r}"
                    )
                if len(key) == 1:
                    key = _character_to_key(key)

                yield Binding(
                    key=key,
                    action=binding.action,
                    description=binding.description,
                    show=bool(binding.description and binding.show),
                    key_display=binding.key_display,
                    priority=binding.priority,
                    tooltip=binding.tooltip,
                    id=binding.id,
                    system=binding.system,
                )


class ActiveBinding(NamedTuple):
    """Information about an active binding (returned from [active_bindings][textual.screen.Screen.active_bindings])."""

    node: DOMNode
    """The node where the binding is defined."""
    binding: Binding
    """The binding information."""
    enabled: bool
    """Is the binding enabled? (enabled bindings are typically rendered dim)"""
    tooltip: str = ""
    """Optional tooltip shown in Footer."""


@rich.repr.auto
class BindingsMap:
    """Manage a set of bindings."""

    def __init__(
        self,
        bindings: Iterable[BindingType] | None = None,
    ) -> None:
        """Initialise a collection of bindings.

        Args:
            bindings: An optional set of initial bindings.

        Note:
            The iterable of bindings can contain either a `Binding`
            instance, or a tuple of 3 values mapping to the first three
            properties of a `Binding`.
        """

        self.key_to_bindings: dict[str, list[Binding]] = {}
        """Mapping of key (e.g. "ctrl+a") to list of bindings for that key."""

        for binding in Binding.make_bindings(bindings or {}):
            self.key_to_bindings.setdefault(binding.key, []).append(binding)

    def _add_binding(self, binding: Binding) -> None:
        """Add a new binding.

        Args:
            binding: New Binding to add.
        """
        self.key_to_bindings.setdefault(binding.key, []).append(binding)

    def __iter__(self) -> Iterator[tuple[str, Binding]]:
        """Iterating produces a sequence of (KEY, BINDING) tuples."""
        return iter(
            [
                (key, binding)
                for key, bindings in self.key_to_bindings.items()
                for binding in bindings
            ]
        )

    @classmethod
    def from_keys(cls, keys: dict[str, list[Binding]]) -> BindingsMap:
        """Construct a BindingsMap from a dict of keys and bindings.

        Args:
            keys: A dict that maps a key on to a list of `Binding` objects.

        Returns:
            New `BindingsMap`
        """
        bindings = cls()
        bindings.key_to_bindings = keys
        return bindings

    def copy(self) -> BindingsMap:
        """Return a copy of this instance.

        Return:
            New bindings object.
        """
        copy = BindingsMap()
        copy.key_to_bindings = self.key_to_bindings.copy()
        return copy

    def __rich_repr__(self) -> rich.repr.Result:
        yield self.key_to_bindings

    @classmethod
    def merge(cls, bindings: Iterable[BindingsMap]) -> BindingsMap:
        """Merge a bindings.

        Args:
            bindings: A number of bindings.

        Returns:
            New `BindingsMap`.
        """
        keys: dict[str, list[Binding]] = {}
        for _bindings in bindings:
            for key, key_bindings in _bindings.key_to_bindings.items():
                keys.setdefault(key, []).extend(key_bindings)
        return BindingsMap.from_keys(keys)

    def apply_keymap(self, keymap: Keymap) -> KeymapApplyResult:
        """Replace bindings for keys that are present in `keymap`.

        Preserves existing bindings for keys that are not in `keymap`.

        Args:
            keymap: A keymap to overlay.

        Returns:
            KeymapApplyResult: The result of applying the keymap, including any clashed bindings.
        """
        clashed_bindings: set[Binding] = set()
        new_bindings: dict[str, list[Binding]] = {}

        key_to_bindings = list(self.key_to_bindings.items())
        for key, bindings in key_to_bindings:
            for binding in bindings:
                binding_id = binding.id
                if binding_id is None:
                    # Bindings without an ID are irrelevant when applying a keymap
                    continue

                # If the keymap has an override for this binding ID
                if keymap_key_string := keymap.get(binding_id):
                    keymap_keys = keymap_key_string.split(",")

                    # Remove the old binding
                    for key, key_bindings in key_to_bindings:
                        key = key.strip()
                        if any(binding.id == binding_id for binding in key_bindings):
                            if key in self.key_to_bindings:
                                del self.key_to_bindings[key]

                    for keymap_key in keymap_keys:
                        if (
                            keymap_key in self.key_to_bindings
                            or keymap_key in new_bindings
                        ):
                            # The key is already mapped either by default or by the keymap,
                            # so there's a clash unless the existing binding is being rebound
                            # to a different key.
                            clashing_bindings = self.key_to_bindings.get(
                                keymap_key, []
                            ) + new_bindings.get(keymap_key, [])
                            for clashed_binding in clashing_bindings:
                                # If the existing binding is not being rebound, it's a clash
                                if not (
                                    clashed_binding.id
                                    and keymap.get(clashed_binding.id)
                                    != clashed_binding.key
                                ):
                                    clashed_bindings.add(clashed_binding)

                            if keymap_key in self.key_to_bindings:
                                del self.key_to_bindings[keymap_key]

                    for keymap_key in keymap_keys:
                        new_bindings.setdefault(keymap_key, []).append(
                            binding.with_key(key=keymap_key, key_display=None)
                        )

        # Update the key_to_bindings with the new bindings
        self.key_to_bindings.update(new_bindings)
        return KeymapApplyResult(clashed_bindings)

    @property
    def shown_keys(self) -> list[Binding]:
        """A list of bindings for shown keys."""
        keys = [
            binding
            for bindings in self.key_to_bindings.values()
            for binding in bindings
            if binding.show
        ]
        return keys

    def bind(
        self,
        keys: str,
        action: str,
        description: str = "",
        show: bool = True,
        key_display: str | None = None,
        priority: bool = False,
    ) -> None:
        """Bind keys to an action.

        Args:
            keys: The keys to bind. Can be a comma-separated list of keys.
            action: The action to bind the keys to.
            description: An optional description for the binding.
            show: A flag to say if the binding should appear in the footer.
            key_display: Optional string to display in the footer for the key.
            priority: Is this a priority binding, checked form app down to focused widget?
        """
        all_keys = [key.strip() for key in keys.split(",")]
        for key in all_keys:
            self.key_to_bindings.setdefault(key, []).append(
                Binding(
                    key,
                    action,
                    description,
                    show=bool(description and show),
                    key_display=key_display,
                    priority=priority,
                )
            )

    def get_bindings_for_key(self, key: str) -> list[Binding]:
        """Get a list of bindings for a given key.

        Args:
            key: Key to look up.

        Raises:
            NoBinding: If the binding does not exist.

        Returns:
            A list of bindings associated with the key.
        """
        try:
            return self.key_to_bindings[key]
        except KeyError:
            raise NoBinding(f"No binding for {key}") from None


class KeymapApplyResult(NamedTuple):
    """The result of applying a keymap."""

    clashed_bindings: set[Binding]
    """A list of bindings that were clashed and replaced by the keymap."""



from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import Label, ListItem, ListView


class MyListView(ListView):
    """Test child class of a ListView."""

    BINDINGS = [Binding(key="s", action="set", description="Set")]

    def __init__(self, items: int = 0) -> None:
        super().__init__()
        self._items = items
        self.action_fired = False

    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        for n in range(self._items):
            yield ListItem(Label(f"This is item {n}"))

    def action_set(self) -> None:
        self.action_fired = True


class ListViewApp(App[None]):
    """ListView test app."""

    def __init__(self, items: int = 0) -> None:
        super().__init__()
        self._items = items

    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        yield MyListView(self._items)


async def test_empty_inherited_list_view() -> None:
    """An empty self-populating inherited ListView should work as expected."""
    async with ListViewApp().run_test() as pilot:
        await pilot.press("tab")
        assert pilot.app.query_one(MyListView).index is None
        await pilot.press("down")
        assert pilot.app.query_one(MyListView).index is None


async def test_populated_inherited_list_view() -> None:
    """A self-populating inherited ListView should work as normal."""
    async with ListViewApp(30).run_test() as pilot:
        await pilot.press("tab")
        assert pilot.app.query_one(MyListView).index == 0
        await pilot.press("down")
        assert pilot.app.query_one(MyListView).index == 1


async def test_actions_work_when_list_view_empty() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/2265"""
    async with ListViewApp().run_test() as pilot:
        await pilot.press("tab", "s")
        list_view = pilot.app.query_one(MyListView)
        assert list_view.action_fired



import httpx
from rich.text import Text

from textual.app import App, ComposeResult
from textual.containers import VerticalScroll
from textual.widgets import Input, Static


class WeatherApp(App):
    """App to display the current weather."""

    CSS_PATH = "weather.tcss"

    def compose(self) -> ComposeResult:
        yield Input(placeholder="Enter a City")
        with VerticalScroll(id="weather-container"):
            yield Static(id="weather")

    async def on_input_changed(self, message: Input.Changed) -> None:
        """Called when the input changes"""
        await self.update_weather(message.value)

    async def update_weather(self, city: str) -> None:
        """Update the weather for the given city."""
        weather_widget = self.query_one("#weather", Static)
        if city:
            # Query the network API
            url = f"https://wttr.in/{city}"
            async with httpx.AsyncClient() as client:
                response = await client.get(url)
                weather = Text.from_ansi(response.text)
                weather_widget.update(weather)
        else:
            # No city, so just blank out the weather
            weather_widget.update("")


if __name__ == "__main__":
    app = WeatherApp()
    app.run()



from typing import NamedTuple


class Transition(NamedTuple):
    duration: float = 1.0
    easing: str = "linear"
    delay: float = 0.0

    def __str__(self) -> str:
        duration, easing, delay = self
        if delay:
            return f"{duration:.1f}s {easing} {delay:.1f}"
        elif easing != "linear":
            return f"{duration:.1f}s {easing}"
        else:
            return f"{duration:.1f}s"



from __future__ import annotations

from textual import events
from textual.app import App, ComposeResult
from textual.geometry import Offset, Region, Size
from textual.reactive import var
from textual.strip import Strip
from textual.scroll_view import ScrollView

from rich.segment import Segment
from rich.style import Style


class CheckerBoard(ScrollView):
    COMPONENT_CLASSES = {
        "checkerboard--white-square",
        "checkerboard--black-square",
        "checkerboard--cursor-square",
    }

    DEFAULT_CSS = """
    CheckerBoard > .checkerboard--white-square {
        background: #A5BAC9;
    }
    CheckerBoard > .checkerboard--black-square {
        background: #004578;
    }
    CheckerBoard > .checkerboard--cursor-square {
        background: darkred;
    }
    """

    cursor_square = var(Offset(0, 0))

    def __init__(self, board_size: int) -> None:
        super().__init__()
        self.board_size = board_size
        # Each square is 4 rows and 8 columns
        self.virtual_size = Size(board_size * 8, board_size * 4)

    def on_mouse_move(self, event: events.MouseMove) -> None:
        """Called when the user moves the mouse over the widget."""
        mouse_position = event.offset + self.scroll_offset
        self.cursor_square = Offset(mouse_position.x // 8, mouse_position.y // 4)

    def watch_cursor_square(
        self, previous_square: Offset, cursor_square: Offset
    ) -> None:
        """Called when the cursor square changes."""

        def get_square_region(square_offset: Offset) -> Region:
            """Get region relative to widget from square coordinate."""
            x, y = square_offset
            region = Region(x * 8, y * 4, 8, 4)
            # Move the region into the widgets frame of reference
            region = region.translate(-self.scroll_offset)
            return region

        # Refresh the previous cursor square
        self.refresh(get_square_region(previous_square))

        # Refresh the new cursor square
        self.refresh(get_square_region(cursor_square))

    def render_line(self, y: int) -> Strip:
        """Render a line of the widget. y is relative to the top of the widget."""

        scroll_x, scroll_y = self.scroll_offset  # The current scroll position
        y += scroll_y  # The line at the top of the widget is now `scroll_y`, not zero!
        row_index = y // 4  # four lines per row

        white = self.get_component_rich_style("checkerboard--white-square")
        black = self.get_component_rich_style("checkerboard--black-square")
        cursor = self.get_component_rich_style("checkerboard--cursor-square")

        if row_index >= self.board_size:
            return Strip.blank(self.size.width)

        is_odd = row_index % 2

        def get_square_style(column: int, row: int) -> Style:
            """Get the cursor style at the given position on the checkerboard."""
            if self.cursor_square == Offset(column, row):
                square_style = cursor
            else:
                square_style = black if (column + is_odd) % 2 else white
            return square_style

        segments = [
            Segment(" " * 8, get_square_style(column, row_index))
            for column in range(self.board_size)
        ]
        strip = Strip(segments, self.board_size * 8)
        # Crop the strip so that is covers the visible area
        strip = strip.crop(scroll_x, scroll_x + self.size.width)
        return strip


class BoardApp(App):
    def compose(self) -> ComposeResult:
        yield CheckerBoard(100)


if __name__ == "__main__":
    app = BoardApp()
    app.run()



from rich.segment import Segment
from rich.style import Style

from textual.app import App, ComposeResult
from textual.strip import Strip
from textual.widget import Widget


class CheckerBoard(Widget):
    """Render an 8x8 checkerboard."""

    def render_line(self, y: int) -> Strip:
        """Render a line of the widget. y is relative to the top of the widget."""

        row_index = y // 4  # A checkerboard square consists of 4 rows

        if row_index >= 8:  # Generate blank lines when we reach the end
            return Strip.blank(self.size.width)

        is_odd = row_index % 2  # Used to alternate the starting square on each row

        white = Style.parse("on white")  # Get a style object for a white background
        black = Style.parse("on black")  # Get a style object for a black background

        # Generate a list of segments with alternating black and white space characters
        segments = [
            Segment(" " * 8, black if (column + is_odd) % 2 else white)
            for column in range(8)
        ]
        strip = Strip(segments, 8 * 8)
        return strip


class BoardApp(App):
    """A simple app to show our widget."""

    def compose(self) -> ComposeResult:
        yield CheckerBoard()


if __name__ == "__main__":
    app = BoardApp()
    app.run()



from textual.app import App, ComposeResult
from textual.containers import VerticalScroll
from textual.widgets import Label


class LoadingOverlayApp(App[None]):
    CSS = """
    VerticalScroll {
        height: 20;
    }
    
    Toast {
        max-width: 100%;
        width: 70%;  /* We need this to cover the dots of the loading indicator
        so that we don't have flakiness in the tests because the screenshot
        might be taken a couple ms earlier or later. */
    }"""

    def compose(self) -> ComposeResult:
        with VerticalScroll():
            yield Label("another big label\n" * 30)  # Ensure there's a scrollbar.

    def on_mount(self):
        self.notify("This is a big notification.\n" * 10, timeout=10)
        self.query_one(VerticalScroll).loading = True


if __name__ == "__main__":
    LoadingOverlayApp().run()



from textual.app import App
from textual.color import Color
from textual.widgets import Label


class TintApp(App):
    CSS_PATH = "tint.tcss"

    def compose(self):
        color = Color.parse("green")
        for tint_alpha in range(0, 101, 10):
            widget = Label(f"tint: green {tint_alpha}%;")
            widget.styles.tint = color.with_alpha(tint_alpha / 100)  # (1)!
            yield widget


if __name__ == "__main__":
    app = TintApp()
    app.run()



from textual.app import App
from textual.containers import Grid
from textual.widgets import Label


class AllBordersApp(App):
    CSS_PATH = "border_all.tcss"

    def compose(self):
        yield Grid(
            Label("ascii", id="ascii"),
            Label("blank", id="blank"),
            Label("dashed", id="dashed"),
            Label("double", id="double"),
            Label("heavy", id="heavy"),
            Label("hidden/none", id="hidden"),
            Label("hkey", id="hkey"),
            Label("inner", id="inner"),
            Label("outer", id="outer"),
            Label("panel", id="panel"),
            Label("round", id="round"),
            Label("solid", id="solid"),
            Label("tall", id="tall"),
            Label("thick", id="thick"),
            Label("vkey", id="vkey"),
            Label("wide", id="wide"),
        )


if __name__ == "__main__":
    app = AllBordersApp()
    app.run()



from textual.app import App, ComposeResult
from textual.widget import Widget
from textual.widgets import Footer, Switch
from textual.binding import Binding


class MyWidget(Widget, can_focus=True):
    BINDINGS = [
        Binding("space", "app.bell", "Bell (Widget)"),
        Binding("a", "app.notify('a widget')", "widget"),
        Binding("b", "app.notify('b widget')", "widget"),
    ]

    DEFAULT_CSS = """
    MyWidget {
        border: solid green;
        height: 5;
    }
    """


class BindApp(App):
    BINDINGS = [
        Binding("space", "app.bell", "Bell (App)"),
        Binding("c", "app.notify('c app')", "app"),
        Binding("a", "app.notify('a app')", "app"),
        Binding("b", "app.notify('b app')", "app"),
    ]

    def compose(self) -> ComposeResult:
        yield MyWidget()
        yield Switch()
        yield Footer()

    def action_notify(self, msg: str):
        self.notify(msg)


if __name__ == "__main__":
    app = BindApp()
    app.run()



"""See https://github.com/Textualize/textual/issues/2900 for the reason behind these tests."""

from textual.app import App, ComposeResult
from textual.widgets import SelectionList


class SelectionListApp(App[None]):
    """Test selection list application."""

    CSS = """
    OptionList {
        width: 20;
    }
    """

    def compose(self) -> ComposeResult:
        yield SelectionList[int](*[(f"{n} ", n) for n in range(10)])


async def test_over_wide_options() -> None:
    """Options wider than the widget should not be an issue."""
    async with SelectionListApp().run_test() as pilot:
        assert pilot.app.query_one(SelectionList).highlighted == 0
        await pilot.pause()
        assert pilot.app.query_one(SelectionList).highlighted == 0


if __name__ == "__main__":
    SelectionListApp().run()



from textual.app import App
from textual.screen import Screen
from textual.widgets import Header


async def test_screen_title_none_is_ignored():
    class MyScreen(Screen):
        def compose(self):
            yield Header()

    class MyApp(App):
        TITLE = "app title"

        def on_mount(self):
            self.push_screen(MyScreen())

    app = MyApp()
    async with app.run_test():
        assert app.query_one("HeaderTitle").text == "app title"


async def test_screen_title_overrides_app_title():
    class MyScreen(Screen):
        TITLE = "screen title"

        def compose(self):
            yield Header()

    class MyApp(App):
        TITLE = "app title"

        def on_mount(self):
            self.push_screen(MyScreen())

    app = MyApp()
    async with app.run_test():
        assert app.query_one("HeaderTitle").text == "screen title"


async def test_screen_title_reactive_updates_title():
    class MyScreen(Screen):
        TITLE = "screen title"

        def compose(self):
            yield Header()

    class MyApp(App):
        TITLE = "app title"

        def on_mount(self):
            self.push_screen(MyScreen())

    app = MyApp()
    async with app.run_test() as pilot:
        app.screen.title = "new screen title"
        await pilot.pause()
        assert app.query_one("HeaderTitle").text == "new screen title"


async def test_app_title_reactive_does_not_update_title_when_screen_title_is_set():
    class MyScreen(Screen):
        TITLE = "screen title"

        def compose(self):
            yield Header()

    class MyApp(App):
        TITLE = "app title"

        def on_mount(self):
            self.push_screen(MyScreen())

    app = MyApp()
    async with app.run_test() as pilot:
        app.title = "new app title"
        await pilot.pause()
        assert app.query_one("HeaderTitle").text == "screen title"


async def test_screen_sub_title_none_is_ignored():
    class MyScreen(Screen):
        def compose(self):
            yield Header()

    class MyApp(App):
        SUB_TITLE = "app sub-title"

        def on_mount(self):
            self.push_screen(MyScreen())

    app = MyApp()
    async with app.run_test():
        assert app.query_one("HeaderTitle").sub_text == "app sub-title"


async def test_screen_sub_title_overrides_app_sub_title():
    class MyScreen(Screen):
        SUB_TITLE = "screen sub-title"

        def compose(self):
            yield Header()

    class MyApp(App):
        SUB_TITLE = "app sub-title"

        def on_mount(self):
            self.push_screen(MyScreen())

    app = MyApp()
    async with app.run_test():
        assert app.query_one("HeaderTitle").sub_text == "screen sub-title"


async def test_screen_sub_title_reactive_updates_sub_title():
    class MyScreen(Screen):
        SUB_TITLE = "screen sub-title"

        def compose(self):
            yield Header()

    class MyApp(App):
        SUB_TITLE = "app sub-title"

        def on_mount(self):
            self.push_screen(MyScreen())

    app = MyApp()
    async with app.run_test() as pilot:
        app.screen.sub_title = "new screen sub-title"
        await pilot.pause()
        assert app.query_one("HeaderTitle").sub_text == "new screen sub-title"


async def test_app_sub_title_reactive_does_not_update_sub_title_when_screen_sub_title_is_set():
    class MyScreen(Screen):
        SUB_TITLE = "screen sub-title"

        def compose(self):
            yield Header()

    class MyApp(App):
        SUB_TITLE = "app sub-title"

        def on_mount(self):
            self.push_screen(MyScreen())

    app = MyApp()
    async with app.run_test() as pilot:
        app.sub_title = "new app sub-title"
        await pilot.pause()
        assert app.query_one("HeaderTitle").sub_text == "screen sub-title"



import pytest

from textual.app import App, ComposeResult
from textual.containers import Container, ScrollableContainer
from textual.screen import Screen
from textual.widget import Widget
from textual.widgets import Button, Label


class Focusable(Widget, can_focus=True):
    pass


class NonFocusable(Widget, can_focus=False, can_focus_children=False):
    pass


class ChildrenFocusableOnly(Widget, can_focus=False, can_focus_children=True):
    pass


@pytest.fixture
def screen() -> Screen:
    app = App()

    with app._context():
        app.push_screen(Screen())

        screen = app.screen

        # The classes even/odd alternate along the focus chain.
        # The classes in/out identify nested widgets.
        screen._add_children(
            Focusable(id="foo", classes="a"),
            NonFocusable(id="bar"),
            Focusable(Focusable(id="Paul", classes="c"), id="container1", classes="b"),
            NonFocusable(Focusable(id="Jessica", classes="a"), id="container2"),
            Focusable(id="baz", classes="b"),
            ChildrenFocusableOnly(Focusable(id="child", classes="c")),
        )

        return screen


def test_focus_chain():
    app = App()
    with app._context():
        app.push_screen(Screen())

        screen = app.screen

        # Check empty focus chain
        assert not screen.focus_chain

        app.screen._add_children(
            Focusable(id="foo"),
            NonFocusable(id="bar"),
            Focusable(Focusable(id="Paul"), id="container1"),
            NonFocusable(Focusable(id="Jessica"), id="container2"),
            Focusable(id="baz"),
            ChildrenFocusableOnly(Focusable(id="child")),
        )

        focus_chain = [widget.id for widget in screen.focus_chain]
        assert focus_chain == ["foo", "container1", "Paul", "baz", "child"]


def test_allow_focus():
    """Test allow_focus and allow_focus_children are called and the result used."""
    focusable_allow_focus_called = False
    non_focusable_allow_focus_called = False

    class Focusable(Widget, can_focus=False):
        def allow_focus(self) -> bool:
            nonlocal focusable_allow_focus_called
            focusable_allow_focus_called = True
            return True

    class NonFocusable(Container, can_focus=True):
        def allow_focus(self) -> bool:
            nonlocal non_focusable_allow_focus_called
            non_focusable_allow_focus_called = True
            return False

    class FocusableContainer(Container, can_focus_children=False):
        def allow_focus_children(self) -> bool:
            return True

    class NonFocusableContainer(Container, can_focus_children=True):
        def allow_focus_children(self) -> bool:
            return False

    app = App()

    with app._context():
        app.push_screen(Screen())

        app.screen._add_children(
            Focusable(id="foo"),
            NonFocusable(id="bar"),
            FocusableContainer(Button("egg", id="egg")),
            NonFocusableContainer(Button("EGG", id="qux")),
        )
        assert [widget.id for widget in app.screen.focus_chain] == ["foo", "egg"]
        assert focusable_allow_focus_called
        assert non_focusable_allow_focus_called


def test_focus_next_and_previous(screen: Screen):
    assert screen.focus_next().id == "foo"
    assert screen.focus_next().id == "container1"
    assert screen.focus_next().id == "Paul"
    assert screen.focus_next().id == "baz"
    assert screen.focus_next().id == "child"

    assert screen.focus_previous().id == "baz"
    assert screen.focus_previous().id == "Paul"
    assert screen.focus_previous().id == "container1"
    assert screen.focus_previous().id == "foo"


def test_focus_next_wrap_around(screen: Screen):
    """Ensure focusing the next widget wraps around the focus chain."""
    screen.set_focus(screen.query_one("#child"))
    assert screen.focused.id == "child"

    assert screen.focus_next().id == "foo"


def test_focus_previous_wrap_around(screen: Screen):
    """Ensure focusing the previous widget wraps around the focus chain."""
    screen.set_focus(screen.query_one("#foo"))
    assert screen.focused.id == "foo"

    assert screen.focus_previous().id == "child"


def test_wrap_around_selector(screen: Screen):
    """Ensure moving focus in both directions wraps around the focus chain."""
    screen.set_focus(screen.query_one("#foo"))
    assert screen.focused.id == "foo"

    assert screen.focus_previous("#Paul").id == "Paul"
    assert screen.focus_next("#foo").id == "foo"


def test_no_focus_empty_selector(screen: Screen):
    """Ensure focus is cleared when selector matches nothing."""
    assert screen.focus_next("#bananas") is None
    assert screen.focus_previous("#bananas") is None

    screen.set_focus(screen.query_one("#foo"))
    assert screen.focused is not None
    assert screen.focus_next("#bananas") is None
    assert screen.focused is None

    screen.set_focus(screen.query_one("#foo"))
    assert screen.focused is not None
    assert screen.focus_previous("#bananas") is None
    assert screen.focused is None


def test_focus_next_and_previous_with_type_selector(screen: Screen):
    """Move focus with a selector that matches the currently focused node."""
    screen.set_focus(screen.query_one("#Paul"))
    assert screen.focused.id == "Paul"

    assert screen.focus_next(Focusable).id == "baz"
    assert screen.focus_next(Focusable).id == "child"

    assert screen.focus_previous(Focusable).id == "baz"
    assert screen.focus_previous(Focusable).id == "Paul"
    assert screen.focus_previous(Focusable).id == "container1"
    assert screen.focus_previous(Focusable).id == "foo"


def test_focus_next_and_previous_with_str_selector(screen: Screen):
    """Move focus with a selector that matches the currently focused node."""
    screen.set_focus(screen.query_one("#foo"))
    assert screen.focused.id == "foo"

    assert screen.focus_next(".a").id == "foo"
    assert screen.focus_next(".c").id == "Paul"
    assert screen.focus_next(".c").id == "child"

    assert screen.focus_previous(".c").id == "Paul"
    assert screen.focus_previous(".a").id == "foo"


def test_focus_next_and_previous_with_type_selector_without_self():
    """Test moving the focus with a selector that does not match the currently focused node."""
    app = App()
    with app._context():
        app.push_screen(Screen())

        screen = app.screen

        from textual.containers import Horizontal, VerticalScroll
        from textual.widgets import Button, Input, Switch

        screen._add_children(
            VerticalScroll(
                Horizontal(
                    Input(id="w3"),
                    Switch(id="w4"),
                    Input(id="w5"),
                    Button(id="w6"),
                    Switch(id="w7"),
                    id="w2",
                ),
                Horizontal(
                    Button(id="w9"),
                    Switch(id="w10"),
                    Button(id="w11"),
                    Input(id="w12"),
                    Input(id="w13"),
                    id="w8",
                ),
                id="w1",
            )
        )

        screen.set_focus(screen.query_one("#w3"))
        assert screen.focused.id == "w3"

        assert screen.focus_next(Button).id == "w6"
        assert screen.focus_next(Switch).id == "w7"
        assert screen.focus_next(Input).id == "w12"

        assert screen.focus_previous(Button).id == "w11"
        assert screen.focus_previous(Switch).id == "w10"
        assert screen.focus_previous(Button).id == "w9"
        assert screen.focus_previous(Input).id == "w5"


def test_focus_next_and_previous_with_str_selector_without_self(screen: Screen):
    """Test moving the focus with a selector that does not match the currently focused node."""
    screen.set_focus(screen.query_one("#foo"))
    assert screen.focused.id == "foo"

    assert screen.focus_next(".c").id == "Paul"
    assert screen.focus_next(".b").id == "baz"
    assert screen.focus_next(".c").id == "child"

    assert screen.focus_previous(".a").id == "foo"
    assert screen.focus_previous(".a").id == "foo"
    assert screen.focus_previous(".b").id == "baz"


async def test_focus_does_not_move_to_invisible_widgets():
    """Make sure invisible widgets don't get focused by accident.

    This is kind of a regression test for https://github.com/Textualize/textual/issues/3053,
    but not really.
    """

    class MyApp(App):
        CSS = "#inv { visibility: hidden; }"

        def compose(self):
            yield Button("one", id="one")
            yield Button("two", id="inv")
            yield Button("three", id="three")

    app = MyApp()
    async with app.run_test():
        assert app.focused.id == "one"
        assert app.screen.focus_next().id == "three"


async def test_focus_moves_to_visible_widgets_inside_invisible_containers():
    """Regression test for https://github.com/Textualize/textual/issues/3053."""

    class MyApp(App):
        CSS = """
        #inv { visibility: hidden; }
        #three { visibility: visible; }
        """

        def compose(self):
            yield Button(id="one")
            with Container(id="inv"):
                yield Button(id="three")

    app = MyApp()
    async with app.run_test():
        assert app.focused.id == "one"
        assert app.screen.focus_next().id == "three"


async def test_focus_chain_handles_inherited_visibility():
    """Regression test for https://github.com/Textualize/textual/issues/3053

    This is more or less a test for the interactions between #3053 and #3071.
    We want to make sure that the focus chain is computed correctly when going through
    a DOM with containers with all sorts of visibilities set.
    """

    class W(Widget):
        can_focus = True

    w1 = W(id="one")
    c2 = Container(id="two")
    w3 = W(id="three")
    c4 = Container(id="four")
    w5 = W(id="five")
    c6 = Container(id="six")
    w7 = W(id="seven")
    c8 = Container(id="eight")
    w9 = W(id="nine")
    w10 = W(id="ten")
    w11 = W(id="eleven")
    w12 = W(id="twelve")
    w13 = W(id="thirteen")

    class InheritedVisibilityApp(App[None]):
        CSS = """
        #four, #eight, #ten {
            visibility: visible;
        }

        #six, #thirteen {
            visibility: hidden;
        }
        """

        def compose(self):
            yield w1  # visible, inherited
            with c2:  # visible, inherited
                yield w3  # visible, inherited
                with c4:  # visible, set
                    yield w5  # visible, inherited
                    with c6:  # hidden, set
                        yield w7  # hidden, inherited
                        with c8:  # visible, set
                            yield w9  # visible, inherited
                        yield w10  # visible, set
                    yield w11  # visible, inherited
                yield w12  # visible, inherited
            yield w13  # invisible, set

    app = InheritedVisibilityApp()
    async with app.run_test():
        focus_chain = app.screen.focus_chain
        assert focus_chain == [
            w1,
            w3,
            w5,
            w9,
            w10,
            w11,
            w12,
        ]


async def test_mouse_down_gives_focus():
    class MyApp(App):
        AUTO_FOCUS = None

        def compose(self):
            yield Button()

    app = MyApp()
    async with app.run_test() as pilot:
        # Sanity check.
        assert app.focused is None

        await pilot.mouse_down(Button)
        assert isinstance(app.focused, Button)


async def test_mouse_up_does_not_give_focus():
    class MyApp(App):
        AUTO_FOCUS = None

        def compose(self):
            yield Button()

    app = MyApp()
    async with app.run_test() as pilot:
        # Sanity check.
        assert app.focused is None

        await pilot.mouse_up(Button)
        assert app.focused is None


async def test_focus_pseudo_class():
    """Test focus and blue pseudo classes"""

    # https://github.com/Textualize/textual/pull/3645
    class FocusApp(App):
        AUTO_FOCUS = None

        def compose(self) -> ComposeResult:
            yield Button("Hello")

    app = FocusApp()
    async with app.run_test() as pilot:
        button = app.query_one(Button)
        classes = list(button.get_pseudo_classes())
        # Blurred, not focused
        assert "blur" in classes
        assert "focus" not in classes

        # Focus the button
        button.focus()
        await pilot.pause()

        # Focused, not blurred
        classes = list(button.get_pseudo_classes())
        assert "blur" not in classes
        assert "focus" in classes


async def test_get_focusable_widget_at() -> None:
    """Check that clicking a non-focusable widget will focus any (focusable) ancestors."""

    class FocusApp(App):
        AUTO_FOCUS = None

        def compose(self) -> ComposeResult:
            with ScrollableContainer(id="focusable"):
                with Container():
                    yield Label("Foo", id="foo")
                    yield Label("Bar", id="bar")
            yield Label("Egg", id="egg")

    app = FocusApp()
    async with app.run_test() as pilot:
        # Nothing focused
        assert app.screen.focused is None
        # Click foo
        await pilot.click("#foo")
        # Confirm container is focused
        assert app.screen.focused is not None
        assert app.screen.focused.id == "focusable"
        # Reset focus
        app.screen.set_focus(None)
        assert app.screen.focused is None
        # Click bar
        await pilot.click("#bar")
        # Confirm container is focused
        assert app.screen.focused is not None
        assert app.screen.focused.id == "focusable"
        # Reset focus
        app.screen.set_focus(None)
        assert app.screen.focused is None
        # Click egg (outside of focusable widget)
        await pilot.click("#egg")
        # Confirm nothing focused
        assert app.screen.focused is None



from textual.app import App, ComposeResult
from textual.widgets import TextArea

TEXT = """\
# The Wonders of Space Exploration

Space exploration has *always* captured the human imagination.

ダレンバーンズ

\tThisissomelongtextthatshouldfoldcorrectly.

\t\tダレン バーンズ







"""


class TextAreaWrapping(App):
    def compose(self) -> ComposeResult:
        yield TextArea.code_editor(TEXT, language="markdown", theme="monokai", soft_wrap=True)


app = TextAreaWrapping()
if __name__ == "__main__":
    app.run()



from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Label

class StackableScreen(Screen):

    TARGET_DEPTH = 10

    def __init__(self, count:int = TARGET_DEPTH) -> None:
        super().__init__()
        self._number = count

    def compose(self) -> ComposeResult:
        yield Label(f"Screen {self.TARGET_DEPTH - self._number}")

    def on_mount(self) -> None:
        if self._number > 0:
            self.app.push_screen(StackableScreen(self._number - 1))

class NotifyDownScreensApp(App[None]):

    def compose(self) -> ComposeResult:
        yield Label("Base screen")

    def on_mount(self):
        for n in range(10):
            self.notify(str(n))
        self.push_screen(StackableScreen())

if __name__ == "__main__":
    NotifyDownScreensApp().run()



from textual.app import App, ComposeResult
from textual.widgets import TextArea

FAIL_TEXT = "x"


class CodeApp(App):
    def compose(self) -> ComposeResult:
        yield TextArea(FAIL_TEXT, soft_wrap=False, show_line_numbers=True)


if __name__ == "__main__":
    app = CodeApp()
    app.run()



from __future__ import annotations

import dataclasses

import pytest

from textual.app import App, ComposeResult
from textual.events import Paste
from textual.widgets import TextArea
from textual.widgets.text_area import EditHistory, Selection

MAX_CHECKPOINTS = 5
SIMPLE_TEXT = """\
ABCDE
FGHIJ
KLMNO
PQRST
UVWXY
Z
"""


@dataclasses.dataclass
class TimeMockableEditHistory(EditHistory):
    mock_time: float | None = dataclasses.field(default=None, init=False)

    def _get_time(self) -> float:
        """Return the mocked time if it is set, otherwise use default behaviour."""
        if self.mock_time is None:
            return super()._get_time()
        return self.mock_time


class TextAreaApp(App):
    def compose(self) -> ComposeResult:
        text_area = TextArea()
        # Update the history object to a version that supports mocking the time.
        text_area.history = TimeMockableEditHistory(
            max_checkpoints=MAX_CHECKPOINTS,
            checkpoint_timer=2.0,
            checkpoint_max_characters=100,
        )
        self.text_area = text_area
        yield text_area


@pytest.fixture
async def pilot():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        yield pilot


@pytest.fixture
async def text_area(pilot):
    return pilot.app.text_area


async def test_simple_undo_redo():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        text_area.insert("123", (0, 0))

        assert text_area.text == "123"
        text_area.undo()
        assert text_area.text == ""
        text_area.redo()
        assert text_area.text == "123"


async def test_undo_selection_retained():
    # Select a range of text and press backspace.
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        text_area.text = SIMPLE_TEXT
        text_area.selection = Selection((0, 0), (2, 3))
        await pilot.press("backspace")
        assert text_area.text == "NO\nPQRST\nUVWXY\nZ\n"
        assert text_area.selection == Selection.cursor((0, 0))

        # Undo the deletion - the text comes back, and the selection is restored.
        text_area.undo()
        assert text_area.selection == Selection((0, 0), (2, 3))
        assert text_area.text == SIMPLE_TEXT

        # Redo the deletion - the text is gone again. The selection goes to the post-delete location.
        text_area.redo()
        assert text_area.text == "NO\nPQRST\nUVWXY\nZ\n"
        assert text_area.selection == Selection.cursor((0, 0))


async def test_undo_checkpoint_created_on_cursor_move():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        text_area.text = SIMPLE_TEXT
        # Characters are inserted on line 0 and 1.
        checkpoint_one = text_area.text
        checkpoint_one_selection = text_area.selection
        await pilot.press("1")  # Added to initial batch.

        # This cursor movement ensures a new checkpoint is created.
        post_insert_one_location = text_area.selection
        await pilot.press("down")

        checkpoint_two = text_area.text
        checkpoint_two_selection = text_area.selection
        await pilot.press("2")  # Added to new batch.

        checkpoint_three = text_area.text
        checkpoint_three_selection = text_area.selection

        # Going back to checkpoint two
        text_area.undo()
        assert text_area.text == checkpoint_two
        assert text_area.selection == checkpoint_two_selection

        # Back again to checkpoint one (initial state)
        text_area.undo()
        assert text_area.text == checkpoint_one
        assert text_area.selection == checkpoint_one_selection

        # Redo to move forward to checkpoint two.
        text_area.redo()
        assert text_area.text == checkpoint_two
        assert text_area.selection == post_insert_one_location

        # Redo to move forward to checkpoint three.
        text_area.redo()
        assert text_area.text == checkpoint_three
        assert text_area.selection == checkpoint_three_selection


async def test_setting_text_property_resets_history():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        await pilot.press("1")

        # Programmatically setting text, which should invalidate the history
        text = "Hello, world!"
        text_area.text = text

        # The undo doesn't do anything, since we set the `text` property.
        text_area.undo()
        assert text_area.text == text


async def test_edits_batched_by_time():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        # The first "12" is batched since they happen within 2 seconds.
        text_area.history.mock_time = 0
        await pilot.press("1")

        text_area.history.mock_time = 1.0
        await pilot.press("2")

        # Since "3" appears 10 seconds later, it's in a separate batch.
        text_area.history.mock_time += 10.0
        await pilot.press("3")

        assert text_area.text == "123"

        text_area.undo()
        assert text_area.text == "12"

        text_area.undo()
        assert text_area.text == ""


async def test_undo_checkpoint_character_limit_reached():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        await pilot.press("1")
        # Since the insertion below is > 100 characters it goes to a new batch.
        text_area.insert("2" * 120)

        text_area.undo()
        assert text_area.text == "1"
        text_area.undo()
        assert text_area.text == ""


async def test_redo_with_no_undo_is_noop():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        text_area.text = SIMPLE_TEXT
        text_area.redo()
        assert text_area.text == SIMPLE_TEXT


async def test_undo_with_empty_undo_stack_is_noop():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        text_area.text = SIMPLE_TEXT
        text_area.undo()
        assert text_area.text == SIMPLE_TEXT


async def test_redo_stack_cleared_on_edit():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        text_area.text = ""
        await pilot.press("1")
        text_area.history.checkpoint()
        await pilot.press("2")
        text_area.history.checkpoint()
        await pilot.press("3")

        text_area.undo()
        text_area.undo()
        text_area.undo()
        assert text_area.text == ""
        assert text_area.selection == Selection.cursor((0, 0))

        # Redo stack has 3 edits in it now.
        await pilot.press("f")
        assert text_area.text == "f"
        assert text_area.selection == Selection.cursor((0, 1))

        # Redo stack is cleared because of the edit, so redo has no effect.
        text_area.redo()
        assert text_area.text == "f"
        assert text_area.selection == Selection.cursor((0, 1))
        text_area.redo()
        assert text_area.text == "f"
        assert text_area.selection == Selection.cursor((0, 1))


async def test_inserts_not_batched_with_deletes():
    # 3 batches here: __1___  ___________2____________  __3__

    app = TextAreaApp()

    async with app.run_test() as pilot:
        text_area = app.text_area
        await pilot.press(*"123", "backspace", "backspace", *"23")

        assert text_area.text == "123"

        # Undo batch 1: the "23" insertion.
        text_area.undo()
        assert text_area.text == "1"

        # Undo batch 2: the double backspace.
        text_area.undo()
        assert text_area.text == "123"

        # Undo batch 3: the "123" insertion.
        text_area.undo()
        assert text_area.text == ""


async def test_paste_is_an_isolated_batch():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        pilot.app.post_message(Paste("hello "))
        pilot.app.post_message(Paste("world"))
        await pilot.pause()

        assert text_area.text == "hello world"

        await pilot.press("!")

        # The insertion of "!" does not get batched with the paste of "world".
        text_area.undo()
        assert text_area.text == "hello world"

        text_area.undo()
        assert text_area.text == "hello "

        text_area.undo()
        assert text_area.text == ""


async def test_focus_creates_checkpoint():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        await pilot.press(*"123")
        text_area.has_focus = False
        text_area.has_focus = True
        await pilot.press(*"456")
        assert text_area.text == "123456"

        # Since we re-focused, a checkpoint exists between 123 and 456,
        # so when we use undo, only the 456 is removed.
        text_area.undo()
        assert text_area.text == "123"


async def test_undo_redo_deletions_batched():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        text_area.text = SIMPLE_TEXT
        text_area.selection = Selection((0, 2), (1, 2))

        # Perform a single delete of some selected text. It'll live in it's own
        # batch since it's a multi-line operation.
        await pilot.press("backspace")
        checkpoint_one = "ABHIJ\nKLMNO\nPQRST\nUVWXY\nZ\n"
        assert text_area.text == checkpoint_one
        assert text_area.selection == Selection.cursor((0, 2))

        # Pressing backspace a few times to delete more characters.
        await pilot.press("backspace", "backspace", "backspace")
        checkpoint_two = "HIJ\nKLMNO\nPQRST\nUVWXY\nZ\n"
        assert text_area.text == checkpoint_two
        assert text_area.selection == Selection.cursor((0, 0))

        # When we undo, the 3 deletions above should be batched, but not
        # the original deletion since it contains a newline character.
        text_area.undo()
        assert text_area.text == checkpoint_one
        assert text_area.selection == Selection.cursor((0, 2))

        # Undoing again restores us back to our initial text and selection.
        text_area.undo()
        assert text_area.text == SIMPLE_TEXT
        assert text_area.selection == Selection((0, 2), (1, 2))

        # At this point, the undo stack contains two items, so we can redo twice.

        # Redo to go back to checkpoint one.
        text_area.redo()
        assert text_area.text == checkpoint_one
        assert text_area.selection == Selection.cursor((0, 2))

        # Redo again to go back to checkpoint two
        text_area.redo()
        assert text_area.text == checkpoint_two
        assert text_area.selection == Selection.cursor((0, 0))

        # Redo again does nothing.
        text_area.redo()
        assert text_area.text == checkpoint_two
        assert text_area.selection == Selection.cursor((0, 0))


async def test_max_checkpoints():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        assert len(text_area.history.undo_stack) == 0
        for index in range(MAX_CHECKPOINTS):
            # Press enter since that will ensure a checkpoint is created.
            await pilot.press("enter")

        assert len(text_area.history.undo_stack) == MAX_CHECKPOINTS
        await pilot.press("enter")
        # Ensure we don't go over the limit.
        assert len(text_area.history.undo_stack) == MAX_CHECKPOINTS


async def test_redo_stack():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        assert len(text_area.history.redo_stack) == 0
        await pilot.press("enter")
        await pilot.press(*"123")
        assert len(text_area.history.undo_stack) == 2
        assert len(text_area.history.redo_stack) == 0
        text_area.undo()
        assert len(text_area.history.undo_stack) == 1
        assert len(text_area.history.redo_stack) == 1
        text_area.undo()
        assert len(text_area.history.undo_stack) == 0
        assert len(text_area.history.redo_stack) == 2
        text_area.redo()
        assert len(text_area.history.undo_stack) == 1
        assert len(text_area.history.redo_stack) == 1
        text_area.redo()
        assert len(text_area.history.undo_stack) == 2
        assert len(text_area.history.redo_stack) == 0


async def test_backward_selection_undo_redo():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.text_area
        # Failed prior to https://github.com/Textualize/textual/pull/4352
        text_area.text = SIMPLE_TEXT
        text_area.selection = Selection((3, 2), (0, 0))

        await pilot.press("a")

        text_area.undo()
        await pilot.press("down", "down", "down", "down")

        assert text_area.text == SIMPLE_TEXT



from typing import Sequence

import pytest

from textual._immutable_sequence_view import ImmutableSequenceView


def wrap(source: Sequence[int]) -> ImmutableSequenceView[int]:
    """Wrap a sequence of integers inside an immutable sequence view."""
    return ImmutableSequenceView[int](source)


def test_empty_immutable_sequence() -> None:
    """An empty immutable sequence should act as anticipated."""
    assert len(wrap([])) == 0
    assert bool(wrap([])) is False
    assert list(wrap([])) == []


def test_non_empty_immutable_sequence() -> None:
    """A non-empty immutable sequence should act as anticipated."""
    assert len(wrap([0])) == 1
    assert bool(wrap([0])) is True
    assert list(wrap([0])) == [0]


def test_no_assign_to_immutable_sequence() -> None:
    """It should not be possible to assign into an immutable sequence."""
    tester = wrap([1, 2, 3, 4, 5])
    with pytest.raises(TypeError):
        tester[0] = 23
    with pytest.raises(TypeError):
        tester[0:3] = 23


def test_no_del_from_iummutable_sequence() -> None:
    """It should not be possible delete an item from an immutable sequence."""
    tester = wrap([1, 2, 3, 4, 5])
    with pytest.raises(TypeError):
        del tester[0]


def test_get_item_from_immutable_sequence() -> None:
    """It should be possible to get an item from an immutable sequence."""
    assert wrap(range(10))[0] == 0
    assert wrap(range(10))[-1] == 9


def test_get_slice_from_immutable_sequence() -> None:
    """It should be possible to get a slice from an immutable sequence."""
    assert list(wrap(range(10))[0:2]) == [0, 1]
    assert list(wrap(range(10))[0:-1]) == [0, 1, 2, 3, 4, 5, 6, 7, 8]


def test_immutable_sequence_contains() -> None:
    """It should be possible to see if an immutable sequence contains a value."""
    tester = wrap([1, 2, 3, 4, 5])
    assert 1 in tester
    assert 11 not in tester


def test_immutable_sequence_index() -> None:
    tester = wrap([1, 2, 3, 4, 5])
    assert tester.index(1) == 0
    with pytest.raises(ValueError):
        _ = tester.index(11)


def test_reverse_immutable_sequence() -> None:
    assert list(reversed(wrap([1, 2]))) == [2, 1]



from __future__ import annotations

from typing import Any

from textual import on
from textual.app import App, ComposeResult
from textual.widgets import Tree


class TreeApp(App[None]):
    """Test tree app."""

    def __init__(self, disabled: bool, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.disabled = disabled
        self.messages: list[tuple[str, str]] = []

    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        yield Tree("Root", disabled=self.disabled, id="test-tree")

    def on_mount(self) -> None:
        self.query_one(Tree).root.add("Child")
        self.query_one(Tree).focus()

    def record(
        self,
        event: (
            Tree.NodeSelected[None]
            | Tree.NodeExpanded[None]
            | Tree.NodeCollapsed[None]
            | Tree.NodeHighlighted[None]
        ),
    ) -> None:
        self.messages.append(
            (event.__class__.__name__, event.node.tree.id or "Unknown")
        )

    @on(Tree.NodeSelected)
    def node_selected(self, event: Tree.NodeSelected[None]) -> None:
        self.record(event)

    @on(Tree.NodeExpanded)
    def node_expanded(self, event: Tree.NodeExpanded[None]) -> None:
        self.record(event)

    @on(Tree.NodeCollapsed)
    def node_collapsed(self, event: Tree.NodeCollapsed[None]) -> None:
        self.record(event)

    @on(Tree.NodeHighlighted)
    def node_highlighted(self, event: Tree.NodeHighlighted[None]) -> None:
        self.record(event)


async def test_creating_disabled_tree():
    """Mounting a disabled `Tree` should result in the base `Widget`
    having a `disabled` property equal to `True`"""
    app = TreeApp(disabled=True)
    async with app.run_test() as pilot:
        tree = app.query_one(Tree)
        assert not tree.focusable
        assert tree.disabled
        assert tree.cursor_line == 0
        await pilot.click("#test-tree")
        await pilot.pause()
        await pilot.press("down")
        await pilot.pause()
        assert tree.cursor_line == 0


async def test_creating_enabled_tree():
    """Mounting an enabled `Tree` should result in the base `Widget`
    having a `disabled` property equal to `False`"""
    app = TreeApp(disabled=False)
    async with app.run_test() as pilot:
        tree = app.query_one(Tree)
        assert tree.focusable
        assert not tree.disabled
        assert tree.cursor_line == 0
        await pilot.click("#test-tree")
        await pilot.pause()
        await pilot.press("down")
        await pilot.pause()
        assert tree.cursor_line == 1


async def test_disabled_tree_node_selected_message() -> None:
    """Clicking the root node disclosure triangle on a disabled tree
    should result in no messages being emitted."""
    app = TreeApp(disabled=True)
    async with app.run_test() as pilot:
        tree = app.query_one(Tree)
        # try clicking on a disabled tree
        await pilot.click("#test-tree")
        await pilot.pause()
        assert pilot.app.messages == [("NodeHighlighted", "test-tree")]
        # make sure messages DO flow after enabling a disabled tree
        tree.disabled = False
        await pilot.click("#test-tree")
        await pilot.pause()
        print(pilot.app.messages)
        assert pilot.app.messages == [
            ("NodeHighlighted", "test-tree"),
            ("NodeExpanded", "test-tree"),
        ]


async def test_enabled_tree_node_selected_message() -> None:
    """Clicking the root node disclosure triangle on an enabled tree
    should result in an `NodeExpanded` message being emitted."""
    app = TreeApp(disabled=False)
    async with app.run_test() as pilot:
        tree = app.query_one(Tree)
        # try clicking on an enabled tree
        await pilot.click("#test-tree")
        await pilot.pause()
        print(pilot.app.messages)
        assert pilot.app.messages == [
            ("NodeHighlighted", "test-tree"),
            ("NodeExpanded", "test-tree"),
        ]
        tree.disabled = True
        # make sure messages DO NOT flow after disabling an enabled tree
        app.messages = []
        await pilot.click("#test-tree")
        await pilot.pause()
        assert not pilot.app.messages



from typing import Union
from unittest import mock

from textual import on
from textual.app import App
from textual.command import CommandPalette, Hit, Hits, Provider

CommandPaletteEvent = Union[
    CommandPalette.Opened, CommandPalette.Closed, CommandPalette.OptionHighlighted
]


class SimpleSource(Provider):
    async def search(self, query: str) -> Hits:
        def goes_nowhere_does_nothing() -> None:
            pass

        yield Hit(1, query, goes_nowhere_does_nothing, query)


class AppWithActiveCommandPalette(App[None]):
    COMMANDS = {SimpleSource}

    def __init__(self) -> None:
        super().__init__()
        self.events: list[CommandPaletteEvent] = []

    def on_mount(self) -> None:
        self.action_command_palette()

    @on(CommandPalette.Opened)
    @on(CommandPalette.Closed)
    @on(CommandPalette.OptionHighlighted)
    def record_event(
        self,
        event: CommandPaletteEvent,
    ) -> None:
        self.events.append(event)


async def test_command_palette_opened_event():
    app = AppWithActiveCommandPalette()
    async with app.run_test():
        assert app.events == [CommandPalette.Opened()]


async def test_command_palette_closed_event():
    app = AppWithActiveCommandPalette()
    async with app.run_test() as pilot:
        await pilot.press("escape")
        assert app.events == [CommandPalette.Opened(), CommandPalette.Closed(False)]


async def test_command_palette_closed_event_value():
    app = AppWithActiveCommandPalette()
    async with app.run_test() as pilot:
        await pilot.press("a")
        await pilot.press("down")
        await pilot.press("enter")
        assert app.events == [
            CommandPalette.Opened(),
            CommandPalette.OptionHighlighted(mock.ANY),
            CommandPalette.Closed(True),
        ]


async def test_command_palette_option_highlighted_event():
    app = AppWithActiveCommandPalette()
    async with app.run_test() as pilot:
        await pilot.press("a")
        await pilot.press("down")
        assert app.events == [
            CommandPalette.Opened(),
            CommandPalette.OptionHighlighted(mock.ANY),
        ]



from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Placeholder


class Header(Placeholder):  # (1)!
    pass


class Footer(Placeholder):  # (2)!
    pass


class TweetScreen(Screen):
    def compose(self) -> ComposeResult:
        yield Header(id="Header")  # (3)!
        yield Footer(id="Footer")  # (4)!


class LayoutApp(App):
    def on_mount(self) -> None:
        self.push_screen(TweetScreen())


if __name__ == "__main__":
    app = LayoutApp()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Input


class BlurApp(App):
    BINDINGS = [("f3", "disable")]

    def compose(self) -> ComposeResult:
        yield Input()

    def on_ready(self) -> None:
        self.query_one(Input).focus()

    def action_disable(self) -> None:
        self.query_one(Input).disabled = True


if __name__ == "__main__":
    app = BlurApp()
    app.run()



from datetime import datetime

from textual.app import App
from textual.widget import Widget


class Clock(Widget):
    """A clock app."""

    DEFAULT_CSS = """
    Clock {
        content-align: center middle;
    }
    """

    def on_mount(self):
        self.set_interval(1, self.refresh)

    def render(self):
        return datetime.now().strftime("%c")


class ClockApp(App):
    def compose(self):
        yield Clock()


if __name__ == "__main__":
    app = ClockApp()
    app.run()



"""Provides a widget for switching between the display of its immediate children."""

from __future__ import annotations

from typing import Optional

from textual.await_complete import AwaitComplete
from textual.containers import Container
from textual.css.query import NoMatches
from textual.events import Mount
from textual.reactive import reactive
from textual.widget import Widget


class ContentSwitcher(Container):
    """A widget for switching between different children.

    Note:
        All child widgets that are to be switched between need a unique ID.
        Children that have no ID will be hidden and ignored.
    """

    DEFAULT_CSS = """
    ContentSwitcher {
        height: auto;
    }

    """

    current: reactive[str | None] = reactive[Optional[str]](None, init=False)
    """The ID of the currently-displayed widget.

    If set to `None` then no widget is visible.

    Note:
        If set to an unknown ID, this will result in
        [`NoMatches`][textual.css.query.NoMatches] being raised.
    """

    def __init__(
        self,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        initial: str | None = None,
    ) -> None:
        """Initialise the content switching widget.

        Args:
            *children: The widgets to switch between.
            name: The name of the content switcher.
            id: The ID of the content switcher in the DOM.
            classes: The CSS classes of the content switcher.
            disabled: Whether the content switcher is disabled or not.
            initial: The ID of the initial widget to show, ``None`` or empty string for the first tab.

        Note:
            If `initial` is not supplied no children will be shown to start with.
        """
        super().__init__(
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self._initial = initial

    def _on_mount(self, _: Mount) -> None:
        """Perform the initial setup of the widget once the DOM is ready."""
        initial = self._initial
        with self.app.batch_update():
            for child in self.children:
                child.display = bool(initial) and child.id == initial
        self._reactive_current = initial

    @property
    def visible_content(self) -> Widget | None:
        """A reference to the currently-visible widget.

        `None` if nothing is visible.
        """
        return self.get_child_by_id(self.current) if self.current is not None else None

    def watch_current(self, old: str | None, new: str | None) -> None:
        """React to the current visible child choice being changed.

        Args:
            old: The old widget ID (or `None` if there was no widget).
            new: The new widget ID (or `None` if nothing should be shown).
        """
        with self.app.batch_update():
            if old:
                try:
                    self.get_child_by_id(old).display = False
                except NoMatches:
                    pass
            if new:
                self.get_child_by_id(new).display = True

    def add_content(
        self, widget: Widget, *, id: str | None = None, set_current: bool = False
    ) -> AwaitComplete:
        """Add new content to the `ContentSwitcher`.

        Args:
            widget: A Widget to add.
            id: ID for the widget, or `None` if the widget already has an ID.
            set_current: Set the new widget as current (which will cause it to display).

        Returns:
            An awaitable to wait for the new content to be mounted.
        """
        if id is not None and widget.id != id:
            widget.id = id

        if not widget.id:
            raise ValueError(
                "Widget must have an ID (or set id parameter when calling add_content)"
            )

        async def _add_content() -> None:
            """Add new widget and potentially change the current widget."""
            widget.display = False
            with self.app.batch_update():
                await self.mount(widget)
                if set_current:
                    self.current = widget.id

        return AwaitComplete(_add_content())



from __future__ import annotations

from typing import Any

import pytest

from textual.actions import ActionError, parse


@pytest.mark.parametrize(
    ("action_string", "expected_namespace", "expected_name", "expected_arguments"),
    [
        ("spam", "", "spam", ()),
        ("hypothetical_action()", "", "hypothetical_action", ()),
        ("another_action(1)", "", "another_action", (1,)),
        ("foo(True, False)", "", "foo", (True, False)),
        ("foo.bar.baz(3, 3.15, 'Python')", "foo.bar", "baz", (3, 3.15, "Python")),
        ("m1234.n5678(None, [1, 2])", "m1234", "n5678", (None, [1, 2])),
    ],
)
def test_parse_action(
    action_string: str,
    expected_namespace: str,
    expected_name: str,
    expected_arguments: tuple[Any],
) -> None:
    namespace, action_name, action_arguments = parse(action_string)
    assert namespace == expected_namespace
    assert action_name == expected_name
    assert action_arguments == expected_arguments


@pytest.mark.parametrize(
    ("action_string", "expected_arguments"),
    [
        ("f()", ()),
        ("f(())", ((),)),
        ("f((1, 2, 3))", ((1, 2, 3),)),
        ("f((1, 2, 3), (1, 2, 3))", ((1, 2, 3), (1, 2, 3))),
        ("f(((1, 2), (), None), 0)", (((1, 2), (), None), 0)),
        ("f((((((1))))))", (1,)),
        ("f(((((((((1, 2)))))))))", ((1, 2),)),
        ("f((1, 2), (3, 4))", ((1, 2), (3, 4))),
        ("f((((((1, 2), (3, 4))))))", (((1, 2), (3, 4)),)),
    ],
)
def test_nested_and_convoluted_tuple_arguments(
    action_string: str, expected_arguments: tuple[Any]
) -> None:
    """Test that tuple arguments are parsed correctly."""
    _, _, args = parse(action_string)
    assert args == expected_arguments


@pytest.mark.parametrize(
    ["action_string", "expected_arguments"],
    [
        ("f('')", ("",)),
        ('f("")', ("",)),
        ("f('''''')", ("",)),
        ('f("""""")', ("",)),
        ("f('(')", ("(",)),
        ("f(')')", (")",)),  # Regression test for #2088
        ("f('f()')", ("f()",)),
    ],
)
def test_parse_action_nested_special_character_arguments(
    action_string: str, expected_arguments: tuple[Any]
) -> None:
    """Test that special characters nested in strings are handled correctly.

    See also: https://github.com/Textualize/textual/issues/2088
    """
    _, _, args = parse(action_string)
    assert args == expected_arguments


@pytest.mark.parametrize(
    "action_string",
    [
        "foo(,,,,,)",
        "bar(1 2 3 4 5)",
        "baz.spam(Tru, Fals, in)",
        "ham(not)",
        "cheese((((()",
    ],
)
def test_parse_action_raises_error(action_string: str) -> None:
    with pytest.raises(ActionError):
        parse(action_string)



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from functools import partial
from typing import TYPE_CHECKING, Iterable

import rich.repr

from textual.css._help_renderables import HelpText
from textual.css.styles import Styles
from textual.css.tokenize import Token
from textual.css.types import Specificity3

if TYPE_CHECKING:
    from typing import Callable

    from typing_extensions import Self

    from textual.dom import DOMNode


class SelectorType(Enum):
    """Type of selector."""

    UNIVERSAL = 1
    """i.e. * operator"""
    TYPE = 2
    """A CSS type, e.g  Label"""
    CLASS = 3
    """CSS class, e.g. .loaded"""
    ID = 4
    """CSS ID, e.g. #main"""
    NESTED = 5
    """Placeholder for nesting operator, i.e &"""


class CombinatorType(Enum):
    """Type of combinator."""

    SAME = 1
    """Selector is combined with previous selector"""
    DESCENDENT = 2
    """Selector is a descendant of the previous selector"""
    CHILD = 3
    """Selector is an immediate child of the previous selector"""


def _check_universal(name: str, node: DOMNode) -> bool:
    """Check node matches universal selector.

    Args:
        name: Selector name.
        node: A DOM node.

    Returns:
        `True` if the selector matches.
    """
    return not node.has_class("-textual-system")


def _check_type(name: str, node: DOMNode) -> bool:
    """Check node matches a type selector.

    Args:
        name: Selector name.
        node: A DOM node.

    Returns:
        `True` if the selector matches.
    """
    return name in node._css_type_names


def _check_class(name: str, node: DOMNode) -> bool:
    """Check node matches a class selector.

    Args:
        name: Selector name.
        node: A DOM node.

    Returns:
        `True` if the selector matches.
    """
    return name in node._classes


def _check_id(name: str, node: DOMNode) -> bool:
    """Check node matches an ID selector.

    Args:
        name: Selector name.
        node: A DOM node.

    Returns:
        `True` if the selector matches.
    """
    return node.id == name


_CHECKS = {
    SelectorType.UNIVERSAL: _check_universal,
    SelectorType.TYPE: _check_type,
    SelectorType.CLASS: _check_class,
    SelectorType.ID: _check_id,
    SelectorType.NESTED: _check_universal,
}


@dataclass
class Selector:
    """Represents a CSS selector.

    Some examples of selectors:

    *
    Header.title
    App > Content
    """

    name: str
    combinator: CombinatorType = CombinatorType.DESCENDENT
    type: SelectorType = SelectorType.TYPE
    pseudo_classes: set[str] = field(default_factory=set)
    specificity: Specificity3 = field(default_factory=lambda: (0, 0, 0))
    advance: int = 1

    def __post_init__(self) -> None:
        self._check: Callable[[DOMNode], bool] = partial(_CHECKS[self.type], self.name)

    @property
    def css(self) -> str:
        """Rebuilds the selector as it would appear in CSS."""
        pseudo_suffix = "".join(f":{name}" for name in sorted(self.pseudo_classes))
        if self.type == SelectorType.UNIVERSAL:
            return "*"
        elif self.type == SelectorType.TYPE:
            return f"{self.name}{pseudo_suffix}"
        elif self.type == SelectorType.CLASS:
            return f".{self.name}{pseudo_suffix}"
        else:
            return f"#{self.name}{pseudo_suffix}"

    def _add_pseudo_class(self, pseudo_class: str) -> None:
        """Adds a pseudo class and updates specificity.

        Args:
            pseudo_class: Name of pseudo class.
        """
        self.pseudo_classes.add(pseudo_class)
        specificity1, specificity2, specificity3 = self.specificity
        self.specificity = (specificity1, specificity2 + 1, specificity3)

    def check(self, node: DOMNode) -> bool:
        """Check if a given node matches the selector.

        Args:
            node: A DOM node.

        Returns:
            True if the selector matches, otherwise False.
        """
        return self._check(node) and (
            node.has_pseudo_classes(self.pseudo_classes)
            if self.pseudo_classes
            else True
        )


@dataclass
class Declaration:
    """A single CSS declaration (not yet processed)."""

    token: Token
    name: str
    tokens: list[Token] = field(default_factory=list)


@rich.repr.auto(angular=True)
@dataclass
class SelectorSet:
    """A set of selectors associated with a rule set."""

    selectors: list[Selector] = field(default_factory=list)
    specificity: Specificity3 = (0, 0, 0)

    def __post_init__(self) -> None:
        SAME = CombinatorType.SAME
        for selector, next_selector in zip(self.selectors, self.selectors[1:]):
            selector.advance = int(next_selector.combinator != SAME)

    @property
    def css(self) -> str:
        return RuleSet._selector_to_css(self.selectors)

    @property
    def is_simple(self) -> bool:
        """Are all the selectors simple (i.e. only dependent on static DOM state)."""
        simple_types = {SelectorType.ID, SelectorType.TYPE}
        return all(
            (selector.type in simple_types and not selector.pseudo_classes)
            for selector in self.selectors
        )

    def __rich_repr__(self) -> rich.repr.Result:
        selectors = RuleSet._selector_to_css(self.selectors)
        yield selectors
        yield None, self.specificity

    def _total_specificity(self) -> Self:
        """Calculate total specificity of selectors.

        Returns:
            Self.
        """
        id_total = class_total = type_total = 0
        for selector in self.selectors:
            _id, _class, _type = selector.specificity
            id_total += _id
            class_total += _class
            type_total += _type
        self.specificity = (id_total, class_total, type_total)
        return self

    @classmethod
    def from_selectors(cls, selectors: list[list[Selector]]) -> Iterable[SelectorSet]:
        for selector_list in selectors:
            id_total = class_total = type_total = 0
            for selector in selector_list:
                _id, _class, _type = selector.specificity
                id_total += _id
                class_total += _class
                type_total += _type
            yield SelectorSet(selector_list, (id_total, class_total, type_total))


@dataclass
class RuleSet:
    selector_set: list[SelectorSet] = field(default_factory=list)
    styles: Styles = field(default_factory=Styles)
    errors: list[tuple[Token, str | HelpText]] = field(default_factory=list)

    is_default_rules: bool = False
    tie_breaker: int = 0
    selector_names: set[str] = field(default_factory=set)
    pseudo_classes: set[str] = field(default_factory=set)

    def __hash__(self):
        return id(self)

    @classmethod
    def _selector_to_css(cls, selectors: list[Selector]) -> str:
        tokens: list[str] = []
        for selector in selectors:
            if selector.combinator == CombinatorType.DESCENDENT:
                tokens.append(" ")
            elif selector.combinator == CombinatorType.CHILD:
                tokens.append(" > ")
            tokens.append(selector.css)

        return "".join(tokens).strip()

    @property
    def selectors(self):
        return ", ".join(
            self._selector_to_css(selector_set.selectors)
            for selector_set in self.selector_set
        )

    @property
    def css(self) -> str:
        """Generate the CSS this RuleSet

        Returns:
            A string containing CSS code.
        """
        declarations = "\n".join(f"    {line}" for line in self.styles.css_lines)
        css = f"{self.selectors} {{\n{declarations}\n}}"
        return css

    def _post_parse(self) -> None:
        """Called after the RuleSet is parsed."""
        # Build a set of the class names that have been updated

        class_type = SelectorType.CLASS
        id_type = SelectorType.ID
        type_type = SelectorType.TYPE
        universal_type = SelectorType.UNIVERSAL

        add_selector = self.selector_names.add
        add_pseudo_classes = self.pseudo_classes.update

        for selector_set in self.selector_set:
            for selector in selector_set.selectors:
                add_pseudo_classes(selector.pseudo_classes)

            selector = selector_set.selectors[-1]
            selector_type = selector.type
            if selector_type == universal_type:
                add_selector("*")
            elif selector_type == type_type:
                add_selector(selector.name)
            elif selector_type == class_type:
                add_selector(f".{selector.name}")
            elif selector_type == id_type:
                add_selector(f"#{selector.name}")



from textual.app import App, ComposeResult
from textual.widgets import Log
from textual.containers import Horizontal

TEXT = (
    """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
I will permit it to pass over me and through me.
And when it has gone past, I will turn the inner eye to see its path.
Where the fear has gone there will be nothing. Only I will remain.
"""
    * 20
)


class LogApp(App):
    def compose(self) -> ComposeResult:
        with Horizontal():
            yield Log(id="log1", auto_scroll=False)
            yield Log(id="log2", auto_scroll=True)
            yield Log(id="log3")
            yield Log(id="log4", max_lines=6)

    def on_ready(self) -> None:
        self.query_one("#log1", Log).write_line(TEXT)
        self.query_one("#log2", Log).write_line(TEXT)
        self.query_one("#log3", Log).write_line(TEXT)
        self.query_one("#log4", Log).write_line(TEXT)

        self.query_one("#log3", Log).clear()
        self.query_one("#log3", Log).write_line("Hello, World")


if __name__ == "__main__":
    app = LogApp()
    app.run()



from __future__ import annotations

from rich.text import Text

from textual.content import Content, Span


def test_blank():
    """Check blank content."""
    blank = Content("")
    assert isinstance(blank, Content)
    assert str(blank) == ""
    assert blank.plain == ""
    assert not blank
    assert blank.markup == ""
    assert len(blank) == 0
    assert blank.spans == []


def test_simple():
    """Check content with simple unstyled text."""
    simple = Content("foo")
    assert isinstance(simple, Content)
    assert str(simple) == "foo"
    assert simple.plain == "foo"
    assert simple  # Not empty is truthy
    assert simple.markup == "foo"
    assert len(simple) == 3
    assert simple.spans == []


def test_constructor():
    content = Content("Hello, World")
    assert content
    assert len(content) == 12
    assert content.cell_length == 12
    assert content.plain == "Hello, World"
    repr(content)


def test_bool():
    assert bool(Content("foo")) is True
    assert bool(Content("")) is False


def test_from_rich_text():
    text = Text.from_markup("[red]Hello[/red] [blue]World[/blue]")
    content = Content.from_rich_text(text)
    assert len(content) == 11
    assert content.plain == "Hello World"
    assert [Span(start=0, end=5, style="red"), Span(start=6, end=11, style="blue")]


def test_styled():
    text = Content.styled("Hello", "red")
    assert text.plain == "Hello"
    assert len(text) == 5
    assert text.cell_length == 5
    assert text._spans == [Span(0, 5, "red")]


def test_getitem():
    content = Content("Hello, world").stylize("blue", 0, 5)
    assert content[0].plain == "H"
    assert content[0]._spans == [Span(0, 1, "blue")]
    assert content[-1].plain == "d"
    assert content[-1]._spans == []
    assert content[:2].plain == "He"
    assert content[:2]._spans == [Span(0, 2, "blue")]


def test_cell_length():
    assert Content("").cell_length == 0
    assert Content("foo").cell_length == 3
    assert Content("💩").cell_length == 2


def test_stylize() -> None:
    """Test the stylize method."""
    foo = Content("foo bar")
    assert foo.spans == []
    red_foo = foo.stylize("red")
    # stylize create a new object
    assert foo.spans == []
    # With no parameters, full string is stylized
    assert red_foo.spans == [Span(0, 7, "red")]
    red_foo = red_foo.stylize("blue", 4, 7)
    # Another span is appended
    assert red_foo.spans == [
        Span(0, 7, "red"),
        Span(4, 7, "blue"),
    ]


def test_stylize_before() -> None:
    """Test the stylize_before method."""
    foo = Content("foo bar")
    assert foo.spans == []
    red_foo = foo.stylize("red")
    # stylize create a new object
    assert foo.spans == []
    # With no parameters, full string is stylized
    assert red_foo.spans == [Span(0, 7, "red")]
    red_foo = red_foo.stylize_before("blue", 4, 7)
    # Another span is appended
    assert red_foo.spans == [
        Span(4, 7, "blue"),
        Span(0, 7, "red"),
    ]


def test_eq() -> None:
    """Test equality."""
    assert Content("foo") == Content("foo")
    assert Content("foo") == "foo"
    assert Content("foo") != Content("bar")
    assert Content("foo") != "bar"


def test_add() -> None:
    """Test addition."""
    # Simple cases
    assert Content("") + Content("") == Content("")
    assert Content("foo") + Content("") == Content("foo")
    # Works with simple strings
    assert Content("foo") + "" == Content("foo")
    assert "" + Content("foo") == Content("foo")

    # Test spans after addition
    content = Content.styled("foo", "red") + " " + Content.styled("bar", "blue")
    assert str(content) == "foo bar"
    assert content.spans == [Span(0, 3, "red"), Span(4, 7, "blue")]
    assert content.cell_length == 7


def test_from_markup():
    """Test simple parsing of Textual markup."""
    content = Content.from_markup("[red]Hello[/red] [blue]World[/blue]")
    assert len(content) == 11
    assert content.plain == "Hello World"
    assert content.spans == [
        Span(start=0, end=5, style="red"),
        Span(start=6, end=11, style="blue"),
    ]


def test_markup():
    """Test markup round trip"""
    content = Content.from_markup("[red]Hello[/red] [blue]World[/blue]")
    assert content.plain == "Hello World"
    assert content.markup == "[red]Hello[/red] [blue]World[/blue]"


def test_join():
    """Test the join method."""

    # Edge cases
    assert Content("").join([]) == ""
    assert Content(".").join([]) == ""
    assert Content("").join(["foo"]) == "foo"
    assert Content(".").join(["foo"]) == "foo"

    # Join strings and Content
    pieces = [Content.styled("foo", "red"), "bar", Content.styled("baz", "blue")]
    content = Content(".").join(pieces)
    assert content.plain == "foo.bar.baz"
    assert content.spans == [Span(0, 3, "red"), Span(8, 11, "blue")]


def test_sort():
    """Test content may be sorted."""
    # functools.total_ordering doing most of the heavy lifting here.
    contents = sorted([Content("foo"), Content("bar"), Content("baz")])
    assert contents[0].plain == "bar"
    assert contents[1].plain == "baz"
    assert contents[2].plain == "foo"


def test_truncate():
    """Test truncated method."""
    content = Content.from_markup("[red]Hello World[/red]")
    # Edge case of 0
    assert content.truncate(0).markup == ""
    # Edge case of 0 wil ellipsis
    assert content.truncate(0, ellipsis=True).markup == ""
    # Edge case of 1
    assert content.truncate(1, ellipsis=True).markup == "[red]…[/red]"
    # Truncate smaller
    assert content.truncate(3).markup == "[red]Hel[/red]"
    # Truncate to same size
    assert content.truncate(11).markup == "[red]Hello World[/red]"
    # Truncate smaller will ellipsis
    assert content.truncate(5, ellipsis=True).markup == "[red]Hell…[/red]"
    # Truncate larger results unchanged
    assert content.truncate(15).markup == "[red]Hello World[/red]"
    # Truncate larger with padding increases size
    assert content.truncate(15, pad=True).markup == "[red]Hello World[/red]    "


def test_assemble():
    """Test Content.assemble constructor."""
    content = Content.assemble(
        "Assemble: ",  # Simple text
        Content.from_markup(
            "pieces of [red]text[/red] or [blue]content[/blue] into "
        ),  # Other Content
        ("a single Content instance", "underline"),  # A tuple of text and a style
    )
    assert (
        content.plain
        == "Assemble: pieces of text or content into a single Content instance"
    )
    print(content.spans)
    assert content.spans == [
        Span(20, 24, style="red"),
        Span(28, 35, style="blue"),
        Span(41, 66, style="underline"),
    ]


def test_escape():
    """Test that escaping the first square bracket."""
    content = Content.from_markup("\\[bold]Not really bold")
    assert content.plain == "[bold]Not really bold"
    assert content.spans == []


def test_strip_control_codes():
    """Test that control codes are removed from content."""
    assert Content("foo\r\nbar").plain == "foo\nbar"
    assert Content.from_markup("foo\r\nbar").plain == "foo\nbar"



from textual.app import App, ComposeResult
from textual.widgets import Static


class ViewportUnits(App):
    CSS = """Static {width: 100vw; height: 100vh; border: solid cyan;} """

    def compose(self) -> ComposeResult:
        yield Static("Hello, world!")


app = ViewportUnits()
if __name__ == '__main__':
    app.run()



PYTHON = """\
import math
from os import path

# I'm a comment :)

string_var = "Hello, world!"
int_var = 42
float_var = 3.14
complex_var = 1 + 2j

list_var = [1, 2, 3, 4, 5]
tuple_var = (1, 2, 3, 4, 5)
set_var = {1, 2, 3, 4, 5}
dict_var = {"a": 1, "b": 2, "c": 3}

def function_no_args():
    return "No arguments"

def function_with_args(a, b):
    return a + b

def function_with_default_args(a=0, b=0):
    return a * b

lambda_func = lambda x: x**2

if int_var == 42:
    print("It's the answer!")
elif int_var < 42:
    print("Less than the answer.")
else:
    print("Greater than the answer.")

for index, value in enumerate(list_var):
    print(f"Index: {index}, Value: {value}")

counter = 0
while counter < 5:
    print(f"Counter value: {counter}")
    counter += 1

squared_numbers = [x**2 for x in range(10) if x % 2 == 0]

try:
    result = 10 / 0
except ZeroDivisionError:
    print("Cannot divide by zero!")
finally:
    print("End of try-except block.")

class Animal:
    def __init__(self, name):
        self.name = name
    
    def speak(self):
        raise NotImplementedError("Subclasses must implement this method.")

class Dog(Animal):
    def speak(self):
        return f"{self.name} says Woof!"

def fibonacci(n):
    a, b = 0, 1
    for _ in range(n):
        yield a
        a, b = b, a + b

for num in fibonacci(5):
    print(num)

with open('test.txt', 'w') as f:
    f.write("Testing with statement.")

@my_decorator
def say_hello():
    print("Hello!")

say_hello()
"""


MARKDOWN = """\
Heading
=======

Sub-heading
-----------

### Heading

#### H4 Heading

##### H5 Heading

###### H6 Heading


Paragraphs are separated
by a blank line.

Two spaces at the end of a line  
produces a line break.

Text attributes _italic_, 
**bold**, `monospace`.

Horizontal rule:

---

Bullet list:

  * apples
  * oranges
  * pears

Numbered list:

  1. lather
  2. rinse
  3. repeat

An [example](http://example.com).

> Markdown uses email-style > characters for blockquoting.
>
> Lorem ipsum

![progress](https://github.com/textualize/rich/raw/master/imgs/progress.gif)


```
a=1
```

```python
import this
```

```somelang
foobar
```

    import this


1. List item

       Code block
"""

YAML = """\
# This is a comment in YAML

# Scalars
string: "Hello, world!"
integer: 42
float: 3.14
boolean: true

# Sequences (Arrays)
fruits:
  - Apple
  - Banana
  - Cherry

# Nested sequences
persons:
  - name: John
    age: 28
    is_student: false
  - name: Jane
    age: 22
    is_student: true

# Mappings (Dictionaries)
address:
  street: 123 Main St
  city: Anytown
  state: CA
  zip: '12345'

# Multiline string
description: |
  This is a multiline
  string in YAML.

# Inline and nested collections
colors: { red: FF0000, green: 00FF00, blue: 0000FF }
"""

TOML = """\
# This is a comment in TOML

string = "Hello, world!"
integer = 42
float = 3.14
boolean = true
datetime = 1979-05-27T07:32:00Z

fruits = ["apple", "banana", "cherry"]

[address]
street = "123 Main St"
city = "Anytown"
state = "CA"
zip = "12345"

[person.john]
name = "John Doe"
age = 28
is_student = false


[[animals]]
name = "Fido"
type = "dog"
"""

SQL = """\
-- This is a comment in SQL

-- Create tables
CREATE TABLE Authors (
    AuthorID INT PRIMARY KEY,
    Name VARCHAR(255) NOT NULL,
    Country VARCHAR(50)
);

CREATE TABLE Books (
    BookID INT PRIMARY KEY,
    Title VARCHAR(255) NOT NULL,
    AuthorID INT,
    PublishedDate DATE,
    FOREIGN KEY (AuthorID) REFERENCES Authors(AuthorID)
);

-- Insert data
INSERT INTO Authors (AuthorID, Name, Country) VALUES (1, 'George Orwell', 'UK');

INSERT INTO Books (BookID, Title, AuthorID, PublishedDate) VALUES (1, '1984', 1, '1949-06-08');

-- Update data
UPDATE Authors SET Country = 'United Kingdom' WHERE Country = 'UK';

-- Select data with JOIN
SELECT Books.Title, Authors.Name 
FROM Books 
JOIN Authors ON Books.AuthorID = Authors.AuthorID;

-- Delete data (commented to preserve data for other examples)
-- DELETE FROM Books WHERE BookID = 1;

-- Alter table structure
ALTER TABLE Authors ADD COLUMN BirthDate DATE;

-- Create index
CREATE INDEX idx_author_name ON Authors(Name);

-- Drop index (commented to avoid actually dropping it)
-- DROP INDEX idx_author_name ON Authors;

-- End of script
"""

CSS = """\
/* This is a comment in CSS */

/* Basic selectors and properties */
body {
    font-family: Arial, sans-serif;
    background-color: #f4f4f4;
    margin: 0;
    padding: 0;
}

/* Class and ID selectors */
.header {
    background-color: #333;
    color: #fff;
    padding: 10px 0;
    text-align: center;
}

#logo {
    font-size: 24px;
    font-weight: bold;
}

/* Descendant and child selectors */
.nav ul {
    list-style-type: none;
    padding: 0;
}

.nav > li {
    display: inline-block;
    margin-right: 10px;
}

/* Pseudo-classes */
a:hover {
    text-decoration: underline;
}

input:focus {
    border-color: #007BFF;
}

/* Media query */
@media (max-width: 768px) {
    body {
        font-size: 16px;
    }
    
    .header {
        padding: 5px 0;
    }
}

/* Keyframes animation */
@keyframes slideIn {
    from {
        transform: translateX(-100%);
    }
    to {
        transform: translateX(0);
    }
}

.slide-in-element {
    animation: slideIn 0.5s forwards;
}
"""

HTML = """\
<!DOCTYPE html>
<html lang="en">

<head>
    <!-- Meta tags -->
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <!-- Title -->
    <title>HTML Test Page</title>
    <!-- Link to CSS -->
    <link rel="stylesheet" href="styles.css">
</head>

<body>
    <!-- Header section -->
    <header class="header">
        <h1 id="logo">HTML Test Page</h1>
    </header>

    <!-- Navigation -->
    <nav class="nav">
        <ul>
            <li><a href="#">Home</a></li>
            <li><a href="#">About</a></li>
            <li><a href="#">Contact</a></li>
        </ul>
    </nav>

    <!-- Main content area -->
    <main>
        <article>
            <h2>Welcome to the Test Page</h2>
            <p>This is a paragraph to test the HTML structure.</p>
            <img src="test-image.jpg" alt="Test Image" width="300">
        </article>
    </main>

    <!-- Form -->
    <section>
        <form action="/submit" method="post">
            <label for="name">Name:</label>
            <input type="text" id="name" name="name">
            <input type="submit" value="Submit">
        </form>
    </section>

    <!-- Footer -->
    <footer>
        <p>&copy; 2023 HTML Test Page</p>
    </footer>

    <!-- Script tag -->
    <script src="scripts.js"></script>
</body>

</html>
"""

JSON = """\
{
    "name": "John Doe",
    "age": 30,
    "isStudent": false,
    "address": {
        "street": "123 Main St",
        "city": "Anytown",
        "state": "CA",
        "zip": "12345"
    },
    "phoneNumbers": [
        {
            "type": "home",
            "number": "555-555-1234"
        },
        {
            "type": "work",
            "number": "555-555-5678"
        }
    ],
    "hobbies": ["reading", "hiking", "swimming"],
    "pets": [
        {
            "type": "dog",
            "name": "Fido"
        },
    ],
    "graduationYear": null
}

"""

REGEX = r"""^abc            # Matches any string that starts with "abc"
abc$            # Matches any string that ends with "abc"
^abc$           # Matches the string "abc" and nothing else
a.b             # Matches any string containing "a", any character, then "b"
a[.]b           # Matches the string "a.b"
a|b             # Matches either "a" or "b"
a{2}            # Matches "aa"
a{2,}           # Matches two or more consecutive "a" characters
a{2,5}          # Matches between 2 and 5 consecutive "a" characters
a?              # Matches "a" or nothing (0 or 1 occurrence of "a")
a*              # Matches zero or more consecutive "a" characters
a+              # Matches one or more consecutive "a" characters
\d              # Matches any digit (equivalent to [0-9])
\D              # Matches any non-digit
\w              # Matches any word character (equivalent to [a-zA-Z0-9_])
\W              # Matches any non-word character
\s              # Matches any whitespace character (spaces, tabs, line breaks)
\S              # Matches any non-whitespace character
(?i)abc         # Case-insensitive match for "abc"
(?:a|b)         # Non-capturing group for either "a" or "b"
(?<=a)b         # Positive lookbehind: matches "b" that is preceded by "a"
(?<!a)b         # Negative lookbehind: matches "b" that is not preceded by "a"
a(?=b)          # Positive lookahead: matches "a" that is followed by "b"
a(?!b)          # Negative lookahead: matches "a" that is not followed by "b"
"""

GO = """\
package main

import (
    "fmt"
    "math"
    "strings"
)

const PI = 3.14159

type Shape interface {
    Area() float64
}

type Circle struct {
    Radius float64
}

func (c Circle) Area() float64 {
    return PI * c.Radius * c.Radius
}

func main() {
    var name string = "John"
    age := 30
    isStudent := true

    fmt.Printf("Hello, %s! You are %d years old.", name, age)

    if age >= 18 && isStudent {
        fmt.Println("You are an adult student.")
    } else if age >= 18 {
        fmt.Println("You are an adult.")
    } else {
        fmt.Println("You are a minor.")
    }

    numbers := []int{1, 2, 3, 4, 5}
    sum := 0
    for _, num := range numbers {
        sum += num
    }
    fmt.Printf("The sum is: %d", sum)

    message := "Hello, World!"
    uppercaseMessage := strings.ToUpper(message)
    fmt.Println(uppercaseMessage)

    circle := Circle{Radius: 5}
    fmt.Printf("Circle area: %.2f", circle.Area())

    result := factorial(5)
    fmt.Printf("Factorial of 5: %d", result)

    defer fmt.Println("Program finished.")

    sqrt := func(x float64) float64 {
        return math.Sqrt(x)
    }
    fmt.Printf("Square root of 16: %.2f", sqrt(16))
}

func factorial(n int) int {
    if n == 0 {
        return 1
    }
    return n * factorial(n-1)
}
"""

JAVASCRIPT = """\
// Variable declarations
const name = "John";
let age = 30;
var isStudent = true;

// Template literals
console.log(`Hello, ${name}! You are ${age} years old.`);

// Conditional statements
if (age >= 18 && isStudent) {
  console.log("You are an adult student.");
} else if (age >= 18) {
  console.log("You are an adult.");
} else {
  console.log("You are a minor.");
}

// Arrays and array methods
const numbers = [1, 2, 3, 4, 5];
const doubledNumbers = numbers.map((num) => num * 2);
console.log("Doubled numbers:", doubledNumbers);

// Objects
const person = {
  firstName: "John",
  lastName: "Doe",
  getFullName() {
    return `${this.firstName} ${this.lastName}`;
  },
};
console.log("Full name:", person.getFullName());

// Classes
class Rectangle {
  constructor(width, height) {
    this.width = width;
    this.height = height;
  }

  getArea() {
    return this.width * this.height;
  }
}
const rectangle = new Rectangle(5, 3);
console.log("Rectangle area:", rectangle.getArea());

// Async/Await and Promises
async function fetchData() {
  try {
    const response = await fetch("https://api.example.com/data");
    const data = await response.json();
    console.log("Fetched data:", data);
  } catch (error) {
    console.error("Error:", error);
  }
}
fetchData();

// Arrow functions
const greet = (name) => {
  console.log(`Hello, ${name}!`);
};
greet("Alice");

// Destructuring assignment
const [a, b, ...rest] = [1, 2, 3, 4, 5];
console.log(a, b, rest);

// Spread operator
const arr1 = [1, 2, 3];
const arr2 = [4, 5, 6];
const combinedArr = [...arr1, ...arr2];
console.log("Combined array:", combinedArr);

// Ternary operator
const message = age >= 18 ? "You are an adult." : "You are a minor.";
console.log(message);
"""

BASH = """\
#!/bin/bash

# Variables
name="John"
age=30
is_student=true

# Printing variables
echo "Hello, $name! You are $age years old."

# Conditional statements
if [[ $age -ge 18 && $is_student == true ]]; then
  echo "You are an adult student."
elif [[ $age -ge 18 ]]; then
  echo "You are an adult."
else
  echo "You are a minor."
fi

# Arrays
numbers=(1 2 3 4 5)
echo "Numbers: ${numbers[@]}"

# Loops
for num in "${numbers[@]}"; do
  echo "Number: $num"
done

# Functions
greet() {
  local name=$1
  echo "Hello, $name!"
}
greet "Alice"

# Command substitution
current_date=$(date +%Y-%m-%d)
echo "Current date: $current_date"

# File operations
touch file.txt
echo "Some content" > file.txt
cat file.txt

# Conditionals with file checks
if [[ -f file.txt ]]; then
  echo "file.txt exists."
else
  echo "file.txt does not exist."
fi

# Case statement
case $age in
  18)
    echo "You are 18 years old."
    ;;
  30)
    echo "You are 30 years old."
    ;;
  *)
    echo "You are neither 18 nor 30 years old."
    ;;
esac

# While loop
counter=0
while [[ $counter -lt 5 ]]; do
  echo "Counter: $counter"
  ((counter++))
done

# Until loop
until [[ $counter -eq 0 ]]; do
  echo "Counter: $counter"
  ((counter--))
done

# Heredoc
cat << EOF
This is a heredoc.
It allows you to write multiple lines of text.
EOF

# Redirection
ls > file_list.txt
grep "file" file_list.txt > filtered_list.txt

# Pipes
cat file_list.txt | wc -l

# Arithmetic operations
result=$((10 + 5))
echo "Result: $result"

# Exporting variables
export DB_PASSWORD="secret"

# Sourcing external files
source config.sh
"""

RUST = """\
use std::collections::HashMap;

// Constants
const PI: f64 = 3.14159;

// Structs
struct Rectangle {
    width: u32,
    height: u32,
}

impl Rectangle {
    fn area(&self) -> u32 {
        self.width * self.height
    }
}

// Enums
enum Result<T, E> {
    Ok(T),
    Err(E),
}

// Functions
fn greet(name: &str) {
    println!("Hello, {}!", name);
}

fn main() {
    // Variables
    let name = "John";
    let mut age = 30;
    let is_student = true;

    // Printing variables
    println!("Hello, {}! You are {} years old.", name, age);

    // Conditional statements
    if age >= 18 && is_student {
        println!("You are an adult student.");
    } else if age >= 18 {
        println!("You are an adult.");
    } else {
        println!("You are a minor.");
    }

    // Arrays
    let numbers = [1, 2, 3, 4, 5];
    println!("Numbers: {:?}", numbers);

    // Vectors
    let mut fruits = vec!["apple", "banana", "orange"];
    fruits.push("grape");
    println!("Fruits: {:?}", fruits);

    // Loops
    for num in &numbers {
        println!("Number: {}", num);
    }

    // Pattern matching
    let result = Result::Ok(42);
    match result {
        Result::Ok(value) => println!("Value: {}", value),
        Result::Err(error) => println!("Error: {:?}", error),
    }

    // Ownership and borrowing
    let s1 = String::from("hello");
    let s2 = s1.clone();
    println!("s1: {}, s2: {}", s1, s2);

    // References
    let rect = Rectangle {
        width: 10,
        height: 20,
    };
    println!("Rectangle area: {}", rect.area());

    // Hash maps
    let mut scores = HashMap::new();
    scores.insert("Alice", 100);
    scores.insert("Bob", 80);
    println!("Alice's score: {}", scores["Alice"]);

    // Closures
    let square = |num: i32| num * num;
    println!("Square of 5: {}", square(5));

    // Traits
    trait Printable {
        fn print(&self);
    }

    impl Printable for Rectangle {
        fn print(&self) {
            println!("Rectangle: width={}, height={}", self.width, self.height);
        }
    }
    rect.print();

    // Modules
    greet("Alice");
}
"""

JAVA = """\
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

// Classes and interfaces
interface Shape {
    double getArea();
}

class Rectangle implements Shape {
    private double width;
    private double height;

    public Rectangle(double width, double height) {
        this.width = width;
        this.height = height;
    }

    @Override
    public double getArea() {
        return width * height;
    }
}

// Enums
enum DaysOfWeek {
    MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY
}

public class Main {
    // Constants
    private static final double PI = 3.14159;

    // Methods
    public static int sum(int a, int b) {
        return a + b;
    }

    public static void main(String[] args) {
        // Variables
        String name = "John";
        int age = 30;
        boolean isStudent = true;

        // Printing variables
        System.out.println("Hello, " + name + "! You are " + age + " years old.");

        // Conditional statements
        if (age >= 18 && isStudent) {
            System.out.println("You are an adult student.");
        } else if (age >= 18) {
            System.out.println("You are an adult.");
        } else {
            System.out.println("You are a minor.");
        }

        // Arrays
        int[] numbers = {1, 2, 3, 4, 5};
        System.out.println("Numbers: " + Arrays.toString(numbers));

        // Lists
        List<String> fruits = new ArrayList<>();
        fruits.add("apple");
        fruits.add("banana");
        fruits.add("orange");
        System.out.println("Fruits: " + fruits);

        // Loops
        for (int num : numbers) {
            System.out.println("Number: " + num);
        }

        // Hash maps
        Map<String, Integer> scores = new HashMap<>();
        scores.put("Alice", 100);
        scores.put("Bob", 80);
        System.out.println("Alice's score: " + scores.get("Alice"));

        // Exception handling
        try {
            int result = 10 / 0;
        } catch (ArithmeticException e) {
            System.out.println("Error: " + e.getMessage());
        }

        // Instantiating objects
        Rectangle rect = new Rectangle(10, 20);
        System.out.println("Rectangle area: " + rect.getArea());

        // Enums
        DaysOfWeek today = DaysOfWeek.MONDAY;
        System.out.println("Today is " + today);

        // Calling methods
        int sum = sum(5, 10);
        System.out.println("Sum: " + sum);

        // Ternary operator
        String message = age >= 18 ? "You are an adult." : "You are a minor.";
        System.out.println(message);
    }
}
"""

XML = """\
<?xml version="1.0" encoding="UTF-8"?>
<!-- This is an example XML document -->
<library>
    <book id="1" genre="fiction">
        <title>The Great Gatsby</title>
        <author>F. Scott Fitzgerald</author>
        <published>1925</published>
        <description><![CDATA[This classic novel explores themes of wealth, society, and the American dream.]]></description>
    </book>
    <book id="2" genre="non-fiction">
        <title>Sapiens: A Brief History of Humankind</title>
        <author>Yuval Noah Harari</author>
        <published>2011</published>
        <description><![CDATA[Explores the history and impact of Homo sapiens.]]></description>
    </book>
    <!-- Another book can be added here -->
</library>
"""

SNIPPETS = {
    "python": PYTHON,
    "markdown": MARKDOWN,
    "yaml": YAML,
    "toml": TOML,
    "sql": SQL,
    "css": CSS,
    "html": HTML,
    "json": JSON,
    "regex": REGEX,
    "go": GO,
    "javascript": JAVASCRIPT,
    "bash": BASH,
    "rust": RUST,
    "java": JAVA,
    "xml": XML,
}



"""

The base class for all messages (including events).
"""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

import rich.repr
from typing_extensions import Self

from textual import _time
from textual._context import active_message_pump
from textual.case import camel_to_snake

if TYPE_CHECKING:
    from textual.dom import DOMNode
    from textual.message_pump import MessagePump


@rich.repr.auto
class Message:
    """Base class for a message."""

    __slots__ = [
        "_sender",
        "time",
        "_forwarded",
        "_no_default_action",
        "_stop_propagation",
        "_prevent",
    ]

    ALLOW_SELECTOR_MATCH: ClassVar[set[str]] = set()
    """Additional attributes that can be used with the [`on` decorator][textual.on].

    These attributes must be widgets.
    """
    bubble: ClassVar[bool] = True  # Message will bubble to parent
    verbose: ClassVar[bool] = False  # Message is verbose
    no_dispatch: ClassVar[bool] = False  # Message may not be handled by client code
    namespace: ClassVar[str] = ""  # Namespace to disambiguate messages
    handler_name: ClassVar[str]
    """Name of the default message handler."""

    def __init__(self) -> None:
        self.__post_init__()

    def __post_init__(self) -> None:
        """Allow dataclasses to initialize the object."""
        self._sender: MessagePump | None = active_message_pump.get(None)
        self.time: float = _time.get_time()
        self._forwarded = False
        self._no_default_action = False
        self._stop_propagation = False
        self._prevent: set[type[Message]] = set()

    def __rich_repr__(self) -> rich.repr.Result:
        yield from ()

    def __init_subclass__(
        cls,
        bubble: bool | None = True,
        verbose: bool = False,
        no_dispatch: bool | None = False,
        namespace: str | None = None,
    ) -> None:
        super().__init_subclass__()
        if bubble is not None:
            cls.bubble = bubble
        cls.verbose = verbose
        if no_dispatch is not None:
            cls.no_dispatch = no_dispatch
        if namespace is not None:
            cls.namespace = namespace
            name = f"{namespace}_{camel_to_snake(cls.__name__)}"
        else:
            # a class defined inside of a function will have a qualified name like func.<locals>.Class,
            # so make sure we only use the actual class name(s)
            qualname = cls.__qualname__.rsplit("<locals>.", 1)[-1]
            # only keep the last two parts of the qualified name of deeply nested classes
            # for backwards compatibility, e.g. A.B.C.D becomes C.D
            namespace = qualname.rsplit(".", 2)[-2:]
            name = "_".join(camel_to_snake(part) for part in namespace)
        cls.handler_name = f"on_{name}"

    @property
    def control(self) -> DOMNode | None:
        """The widget associated with this message, or None by default."""
        return None

    @property
    def is_forwarded(self) -> bool:
        """Has the message been forwarded?"""
        return self._forwarded

    def _set_forwarded(self) -> None:
        """Mark this event as being forwarded."""
        self._forwarded = True

    def set_sender(self, sender: MessagePump) -> Self:
        """Set the sender of the message.

        Args:
            sender: The sender.

        Note:
            When creating a message the sender is automatically set.
            Normally there will be no need for this method to be called.
            This method will be used when strict control is required over
            the sender of a message.

        Returns:
            Self.
        """
        self._sender = sender
        return self

    def can_replace(self, message: "Message") -> bool:
        """Check if another message may supersede this one.

        Args:
            message: Another message.

        Returns:
            True if this message may replace the given message
        """
        return False

    def prevent_default(self, prevent: bool = True) -> Message:
        """Suppress the default action(s). This will prevent handlers in any base classes
        from being called.

        Args:
            prevent: True if the default action should be suppressed,
                or False if the default actions should be performed.
        """
        self._no_default_action = prevent
        return self

    def stop(self, stop: bool = True) -> Message:
        """Stop propagation of the message to parent.

        Args:
            stop: The stop flag.
        """
        self._stop_propagation = stop
        return self

    def _bubble_to(self, widget: MessagePump) -> None:
        """Bubble to a widget (typically the parent).

        Args:
            widget: Target of bubble.
        """
        self._no_default_action = False
        widget.post_message(self)



import os
from pathlib import Path

from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Label

CSS_PATH = (Path(__file__) / "../css_reloading.tcss").resolve()


class BaseScreen(Screen[None]):
    def compose(self) -> ComposeResult:
        yield Label("I am the base screen")


class TopScreen(Screen[None]):
    DEFAULT_CSS = """
    TopScreen {
        opacity: 1;
        background: green 0%;
    }
    """


class MyApp(App[None]):
    CSS_PATH = CSS_PATH

    def on_mount(self) -> None:
        self.push_screen(BaseScreen())
        self.push_screen(TopScreen())


async def test_css_reloading_applies_to_non_top_screen(monkeypatch) -> None:  # type: ignore
    """Regression test for https://github.com/Textualize/textual/issues/3931"""

    monkeypatch.setenv(
        "TEXTUAL", "debug"
    )  # This will make sure we create a file monitor.

    # Write some initial CSS.
    Path(CSS_PATH).write_text(
        """\
Label {
    height: 5;
    border: panel white;
}
"""
    )

    app = MyApp()
    async with app.run_test() as pilot:
        await pilot.pause()
        first_label = pilot.app.screen_stack[-2].query(Label).first()
        # Sanity check.
        assert first_label.styles.height is not None
        assert first_label.styles.height.value == 5

        # Clear the CSS from the file.
        Path(CSS_PATH).write_text("/* This file has no rules intentionally. */\n")
        await pilot.pause()
        await pilot.app._on_css_change()
        # Height should fall back to 1.
        assert first_label.styles.height is not None
        assert first_label.styles.height.value == 1


async def test_css_reloading_file_not_found(monkeypatch, tmp_path):
    """Regression test for https://github.com/Textualize/textual/issues/3996

    Files can become temporarily unavailable during saving on certain environments.
    """
    monkeypatch.setenv("TEXTUAL", "debug")

    css_path = tmp_path / "test_css_reloading_file_not_found.tcss"
    with open(css_path, "w") as css_file:
        css_file.write("#a {color: red;}")

    class TextualApp(App):
        CSS_PATH = css_path

    app = TextualApp()
    async with app.run_test() as pilot:
        await pilot.app._on_css_change()
        os.remove(css_path)
        assert not css_path.exists()
        await pilot.app._on_css_change()



"""Implements a progress bar widget."""

from __future__ import annotations

from typing import Optional

from rich.style import Style

from textual._types import UnusedParameter
from textual.app import ComposeResult, RenderResult
from textual.clock import Clock
from textual.color import Gradient
from textual.eta import ETA
from textual.geometry import clamp
from textual.reactive import reactive
from textual.renderables.bar import Bar as BarRenderable
from textual.widget import Widget
from textual.widgets import Label

UNUSED = UnusedParameter()
"""Sentinel for method signatures."""


class Bar(Widget, can_focus=False):
    """The bar portion of the progress bar."""

    COMPONENT_CLASSES = {"bar--bar", "bar--complete", "bar--indeterminate"}
    """
    The bar sub-widget provides the component classes that follow.

    These component classes let you modify the foreground and background color of the
    bar in its different states.

    | Class | Description |
    | :- | :- |
    | `bar--bar` | Style of the bar (may be used to change the color). |
    | `bar--complete` | Style of the bar when it's complete. |
    | `bar--indeterminate` | Style of the bar when it's in an indeterminate state. |
    """

    DEFAULT_CSS = """
    Bar {
        width: 32;
        height: 1;

        &> .bar--bar {
            color: $primary;
            background: $surface;
        }
        &> .bar--indeterminate {
            color: $error;
            background: $surface;
        }
        &> .bar--complete {
            color: $success;
            background: $surface;
        }
    }
    """

    percentage: reactive[float | None] = reactive[Optional[float]](None)
    """The percentage of progress that has been completed."""

    gradient: reactive[Gradient | None] = reactive(None)
    """An optional gradient."""

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        clock: Clock | None = None,
        gradient: Gradient | None = None,
    ):
        """Create a bar for a [`ProgressBar`][textual.widgets.ProgressBar]."""
        self._clock = (clock or Clock()).clone()
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.set_reactive(Bar.gradient, gradient)

    def _validate_percentage(self, percentage: float | None) -> float | None:
        """Avoid updating the bar, if the percentage increase is too small to render."""
        width = self.size.width * 2
        return (
            None
            if percentage is None
            else (int(percentage * width) / width if width else percentage)
        )

    def watch_percentage(self, percentage: float | None) -> None:
        """Manage the timer that enables the indeterminate bar animation."""
        if percentage is not None:
            self.auto_refresh = None
        else:
            self.auto_refresh = 1 / 15

    def render(self) -> RenderResult:
        """Render the bar with the correct portion filled."""
        if self.percentage is None:
            return self.render_indeterminate()
        else:
            bar_style = (
                self.get_component_rich_style("bar--bar")
                if self.percentage < 1
                else self.get_component_rich_style("bar--complete")
            )
            return BarRenderable(
                highlight_range=(0, self.size.width * self.percentage),
                highlight_style=Style.from_color(bar_style.color),
                background_style=Style.from_color(bar_style.bgcolor),
                gradient=self.gradient,
            )

    def render_indeterminate(self) -> RenderResult:
        """Render a frame of the indeterminate progress bar animation."""
        width = self.size.width
        highlighted_bar_width = 0.25 * width
        # Width used to enable the visual effect of the bar going into the corners.
        total_imaginary_width = width + highlighted_bar_width
        start: float
        end: float
        if self.app.animation_level == "none":
            start = 0
            end = width
        else:
            speed = 30  # Cells per second.
            # Compute the position of the bar.
            start = (speed * self._clock.time) % (2 * total_imaginary_width)
            if start > total_imaginary_width:
                # If the bar is to the right of its width, wrap it back from right to left.
                start = 2 * total_imaginary_width - start  # = (tiw - (start - tiw))
            start -= highlighted_bar_width
            end = start + highlighted_bar_width

        bar_style = self.get_component_rich_style("bar--indeterminate")
        return BarRenderable(
            highlight_range=(max(0, start), min(end, width)),
            highlight_style=Style.from_color(bar_style.color),
            background_style=Style.from_color(bar_style.bgcolor),
        )


class PercentageStatus(Label):
    """A label to display the percentage status of the progress bar."""

    DEFAULT_CSS = """
    PercentageStatus {
        width: 5;
        content-align-horizontal: right;
    }
    """

    percentage: reactive[int | None] = reactive[Optional[int]](None)
    """The percentage of progress that has been completed."""

    def _validate_percentage(self, percentage: float | None) -> int | None:
        return None if percentage is None else round(percentage * 100)

    def render(self) -> RenderResult:
        return "--%" if self.percentage is None else f"{self.percentage}%"


class ETAStatus(Label):
    """A label to display the estimated time until completion of the progress bar."""

    DEFAULT_CSS = """
    ETAStatus {
        width: 9;
        content-align-horizontal: right;
    }
    """
    eta: reactive[float | None] = reactive[Optional[float]](None)
    """Estimated number of seconds till completion, or `None` if no estimate is available."""

    def render(self) -> RenderResult:
        """Render the ETA display."""
        eta = self.eta
        if eta is None:
            return "--:--:--"
        else:
            minutes, seconds = divmod(round(eta), 60)
            hours, minutes = divmod(minutes, 60)
            if hours > 999999:
                return "+999999h"
            elif hours > 99:
                return f"{hours}h"
            else:
                return f"{hours:02}:{minutes:02}:{seconds:02}"


class ProgressBar(Widget, can_focus=False):
    """A progress bar widget."""

    DEFAULT_CSS = """
    ProgressBar {
        width: auto;
        height: 1;
        layout: horizontal;
    }
    """

    progress: reactive[float] = reactive(0.0)
    """The progress so far, in number of steps."""
    total: reactive[float | None] = reactive[Optional[float]](None)
    """The total number of steps associated with this progress bar, when known.

    The value `None` will render an indeterminate progress bar.
    """
    percentage: reactive[float | None] = reactive[Optional[float]](None)
    """The percentage of progress that has been completed.

    The percentage is a value between 0 and 1 and the returned value is only
    `None` if the total progress of the bar hasn't been set yet.

    Example:
        ```py
        progress_bar = ProgressBar()
        print(progress_bar.percentage)  # None
        progress_bar.update(total=100)
        progress_bar.advance(50)
        print(progress_bar.percentage)  # 0.5
        ```
    """
    _display_eta: reactive[int | None] = reactive[Optional[int]](None)

    gradient: reactive[Gradient | None] = reactive(None)
    """Optional gradient object (will replace CSS styling in bar)."""

    def __init__(
        self,
        total: float | None = None,
        *,
        show_bar: bool = True,
        show_percentage: bool = True,
        show_eta: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        clock: Clock | None = None,
        gradient: Gradient | None = None,
    ):
        """Create a Progress Bar widget.

        The progress bar uses "steps" as the measurement unit.

        Example:
            ```py
            class MyApp(App):
                def compose(self):
                    yield ProgressBar(total=100)

                def key_space(self):
                    self.query_one(ProgressBar).advance(5)
            ```

        Args:
            total: The total number of steps in the progress if known.
            show_bar: Whether to show the bar portion of the progress bar.
            show_percentage: Whether to show the percentage status of the bar.
            show_eta: Whether to show the ETA countdown of the progress bar.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
            clock: An optional clock object (leave as default unless testing).
            gradient: An optional Gradient object (will replace CSS styles in the bar).
        """
        self._clock = clock or Clock()
        self._eta = ETA()
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.total = total
        self.show_bar = show_bar
        self.show_percentage = show_percentage
        self.show_eta = show_eta
        self.set_reactive(ProgressBar.gradient, gradient)

    def on_mount(self) -> None:
        self.update()
        self.set_interval(1, self.update)
        self._clock.reset()

    def compose(self) -> ComposeResult:
        if self.show_bar:
            yield (
                Bar(id="bar", clock=self._clock)
                .data_bind(ProgressBar.percentage)
                .data_bind(ProgressBar.gradient)
            )
        if self.show_percentage:
            yield PercentageStatus(id="percentage").data_bind(ProgressBar.percentage)
        if self.show_eta:
            yield ETAStatus(id="eta").data_bind(eta=ProgressBar._display_eta)

    def _validate_total(self, total: float | None) -> float | None:
        """Ensure the total is not negative."""
        if total is None:
            return total
        return max(0, total)

    def _compute_percentage(self) -> float | None:
        """Keep the percentage of progress updated automatically.

        This will report a percentage of `1` if the total is zero.
        """
        if self.total:
            return clamp(self.progress / self.total, 0.0, 1.0)
        elif self.total == 0:
            return 1.0
        return None

    def _watch_progress(self, progress: float) -> None:
        """Perform update when progress is modified."""
        self.update(progress=progress)

    def _watch_total(self, total: float) -> None:
        """Update when the total is modified."""
        self.update(total=total)

    def advance(self, advance: float = 1) -> None:
        """Advance the progress of the progress bar by the given amount.

        Example:
            ```py
            progress_bar.advance(10)  # Advance 10 steps.
            ```

        Args:
            advance: Number of steps to advance progress by.
        """
        self.update(advance=advance)

    def update(
        self,
        *,
        total: None | float | UnusedParameter = UNUSED,
        progress: float | UnusedParameter = UNUSED,
        advance: float | UnusedParameter = UNUSED,
    ) -> None:
        """Update the progress bar with the given options.

        Example:
            ```py
            progress_bar.update(
                total=200,  # Set new total to 200 steps.
                progress=50,  # Set the progress to 50 (out of 200).
            )
            ```

        Args:
            total: New total number of steps.
            progress: Set the progress to the given number of steps.
            advance: Advance the progress by this number of steps.
        """
        current_time = self._clock.time
        if not isinstance(total, UnusedParameter):
            if total is None or total != self.total:
                self._eta.reset()
            self.total = total

        def add_sample() -> None:
            """Add a new sample."""
            if self.progress is not None and self.total:
                self._eta.add_sample(current_time, self.progress / self.total)

        if not isinstance(progress, UnusedParameter):
            self.progress = progress
            add_sample()

        if not isinstance(advance, UnusedParameter):
            self.progress += advance
            add_sample()

        self._display_eta = (
            None if self.total is None else self._eta.get_eta(current_time)
        )



from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import Footer


class FooterApp(App):
    BINDINGS = [
        Binding(key="q", action="quit", description="Quit the app"),
        Binding(
            key="question_mark",
            action="help",
            description="Show help screen",
            key_display="?",
        ),
        Binding(key="delete", action="delete", description="Delete the thing"),
        Binding(key="j", action="down", description="Scroll down", show=False),
    ]

    def compose(self) -> ComposeResult:
        yield Footer()


if __name__ == "__main__":
    app = FooterApp()
    app.run()



"""
This module contains the `Worker` class and related objects.

See the guide for how to use [workers](/guide/workers).

"""

from __future__ import annotations

import asyncio
import enum
import inspect
from contextvars import ContextVar
from threading import Event
from time import monotonic
from typing import (
    TYPE_CHECKING,
    Awaitable,
    Callable,
    Coroutine,
    Generic,
    TypeVar,
    Union,
    cast,
)

import rich.repr
from typing_extensions import TypeAlias

from textual.message import Message

if TYPE_CHECKING:
    from textual.app import App
    from textual.dom import DOMNode


active_worker: ContextVar[Worker] = ContextVar("active_worker")
"""Currently active worker context var."""


class NoActiveWorker(Exception):
    """There is no active worker."""


class WorkerError(Exception):
    """A worker related error."""


class WorkerFailed(WorkerError):
    """The worker raised an exception and did not complete."""

    def __init__(self, error: BaseException) -> None:
        self.error = error
        super().__init__(f"Worker raised exception: {error!r}")


class DeadlockError(WorkerError):
    """The operation would result in a deadlock."""


class WorkerCancelled(WorkerError):
    """The worker was cancelled and did not complete."""


def get_current_worker() -> Worker:
    """Get the currently active worker.

    Raises:
        NoActiveWorker: If there is no active worker.

    Returns:
        A Worker instance.
    """
    try:
        return active_worker.get()
    except LookupError:
        raise NoActiveWorker(
            "There is no active worker in this task or thread."
        ) from None


class WorkerState(enum.Enum):
    """A description of the worker's current state."""

    PENDING = 1
    """Worker is initialized, but not running."""
    RUNNING = 2
    """Worker is running."""
    CANCELLED = 3
    """Worker is not running, and was cancelled."""
    ERROR = 4
    """Worker is not running, and exited with an error."""
    SUCCESS = 5
    """Worker is not running, and completed successfully."""


ResultType = TypeVar("ResultType")


WorkType: TypeAlias = Union[
    Callable[[], Coroutine[None, None, ResultType]],
    Callable[[], ResultType],
    Awaitable[ResultType],
]
"""Type used for [workers](/guide/workers/)."""


class _ReprText:
    """Shim to insert a word into the Worker's repr."""

    def __init__(self, text: str) -> None:
        self.text = text

    def __repr__(self) -> str:
        return self.text


@rich.repr.auto(angular=True)
class Worker(Generic[ResultType]):
    """A class to manage concurrent work (either a task or a thread)."""

    @rich.repr.auto
    class StateChanged(Message, bubble=False, namespace="worker"):
        """The worker state changed."""

        def __init__(self, worker: Worker, state: WorkerState) -> None:
            """Initialize the StateChanged message.

            Args:
                worker: The worker object.
                state: New state.
            """
            self.worker = worker
            self.state = state
            super().__init__()

        def __rich_repr__(self) -> rich.repr.Result:
            yield self.worker
            yield self.state

    def __init__(
        self,
        node: DOMNode,
        work: WorkType,
        *,
        name: str = "",
        group: str = "default",
        description: str = "",
        exit_on_error: bool = True,
        thread: bool = False,
    ) -> None:
        """Initialize a Worker.

        Args:
            node: The widget, screen, or App that initiated the work.
            work: A callable, coroutine, or other awaitable object to run in the worker.
            name: Name of the worker (short string to help identify when debugging).
            group: The worker group.
            description: Description of the worker (longer string with more details).
            exit_on_error: Exit the app if the worker raises an error. Set to `False` to suppress exceptions.
            thread: Mark the worker as a thread worker.
        """
        self._node = node
        self._work = work
        self.name = name
        self.group = group
        self.description = description
        self.exit_on_error = exit_on_error
        self.cancelled_event: Event = Event()
        """A threading event set when the worker is cancelled."""
        self._thread_worker = thread
        self._state = WorkerState.PENDING
        self.state = self._state
        self._error: BaseException | None = None
        self._completed_steps: int = 0
        self._total_steps: int | None = None
        self._cancelled: bool = False
        self._created_time = monotonic()
        self._result: ResultType | None = None
        self._task: asyncio.Task | None = None
        self._node.post_message(self.StateChanged(self, self._state))

    def __rich_repr__(self) -> rich.repr.Result:
        yield _ReprText(self.state.name)
        yield "name", self.name, ""
        yield "group", self.group, "default"
        yield "description", self.description, ""
        yield "progress", round(self.progress, 1), 0.0

    @property
    def node(self) -> DOMNode:
        """The node where this worker was run from."""
        return self._node

    @property
    def state(self) -> WorkerState:
        """The current state of the worker."""
        return self._state

    @state.setter
    def state(self, state: WorkerState) -> None:
        """Set the state, and send a message."""
        changed = state != self._state
        self._state = state
        if changed:
            self._node.post_message(self.StateChanged(self, state))

    @property
    def is_cancelled(self) -> bool:
        """Has the work been cancelled?

        Note that cancelled work may still be running.
        """
        return self._cancelled

    @property
    def is_running(self) -> bool:
        """Is the task running?"""
        return self.state == WorkerState.RUNNING

    @property
    def is_finished(self) -> bool:
        """Has the task finished (cancelled, error, or success)?"""
        return self.state in (
            WorkerState.CANCELLED,
            WorkerState.ERROR,
            WorkerState.SUCCESS,
        )

    @property
    def completed_steps(self) -> int:
        """The number of completed steps."""
        return self._completed_steps

    @property
    def total_steps(self) -> int | None:
        """The number of total steps, or None if indeterminate."""
        return self._total_steps

    @property
    def progress(self) -> float:
        """Progress as a percentage.

        If the total steps is None, then this will return 0. The percentage will be clamped between 0 and 100.
        """
        if not self._total_steps:
            return 0.0
        return max(0, min(100, (self._completed_steps / self._total_steps) * 100.0))

    @property
    def result(self) -> ResultType | None:
        """The result of the worker, or `None` if there is no result."""
        return self._result

    @property
    def error(self) -> BaseException | None:
        """The exception raised by the worker, or `None` if there was no error."""
        return self._error

    def update(
        self, completed_steps: int | None = None, total_steps: int | None = -1
    ) -> None:
        """Update the number of completed steps.

        Args:
            completed_steps: The number of completed seps, or `None` to not change.
            total_steps: The total number of steps, `None` for indeterminate, or -1 to leave unchanged.
        """
        if completed_steps is not None:
            self._completed_steps += completed_steps
        if total_steps != -1:
            self._total_steps = None if total_steps is None else max(0, total_steps)

    def advance(self, steps: int = 1) -> None:
        """Advance the number of completed steps.

        Args:
            steps: Number of steps to advance.
        """
        self._completed_steps += steps

    async def _run_threaded(self) -> ResultType:
        """Run a threaded worker.

        Returns:
            Return value of the work.
        """

        def run_awaitable(work: Awaitable[ResultType]) -> ResultType:
            """Set the active worker and await the awaitable."""

            async def do_work() -> ResultType:
                active_worker.set(self)
                return await work

            return asyncio.run(do_work())

        def run_coroutine(
            work: Callable[[], Coroutine[None, None, ResultType]],
        ) -> ResultType:
            """Set the active worker and await coroutine."""
            return run_awaitable(work())

        def run_callable(work: Callable[[], ResultType]) -> ResultType:
            """Set the active worker, and call the callable."""
            active_worker.set(self)
            return work()

        if (
            inspect.iscoroutinefunction(self._work)
            or hasattr(self._work, "func")
            and inspect.iscoroutinefunction(self._work.func)
        ):
            runner = run_coroutine
        elif inspect.isawaitable(self._work):
            runner = run_awaitable
        elif callable(self._work):
            runner = run_callable
        else:
            raise WorkerError("Unsupported attempt to run a thread worker")

        loop = asyncio.get_running_loop()
        assert loop is not None
        return await loop.run_in_executor(None, runner, self._work)

    async def _run_async(self) -> ResultType:
        """Run an async worker.

        Returns:
            Return value of the work.
        """
        if (
            inspect.iscoroutinefunction(self._work)
            or hasattr(self._work, "func")
            and inspect.iscoroutinefunction(self._work.func)
        ):
            return await self._work()
        elif inspect.isawaitable(self._work):
            return await self._work
        elif callable(self._work):
            raise WorkerError("Request to run a non-async function as an async worker")
        raise WorkerError("Unsupported attempt to run an async worker")

    async def run(self) -> ResultType:
        """Run the work.

        Implement this method in a subclass, or pass a callable to the constructor.

        Returns:
            Return value of the work.
        """
        return await (
            self._run_threaded() if self._thread_worker else self._run_async()
        )

    async def _run(self, app: App) -> None:
        """Run the worker.

        Args:
            app: App instance.
        """
        with app._context():
            active_worker.set(self)

            self.state = WorkerState.RUNNING
            app.log.worker(self)
            try:
                self._result = await self.run()
            except asyncio.CancelledError as error:
                self.state = WorkerState.CANCELLED
                self._error = error
                app.log.worker(self)
            except Exception as error:
                self.state = WorkerState.ERROR
                self._error = error
                app.log.worker(self, "failed", repr(error))
                from rich.traceback import Traceback

                app.log.worker(Traceback())
                if self.exit_on_error:
                    worker_failed = WorkerFailed(self._error)
                    app._handle_exception(worker_failed)
            else:
                self.state = WorkerState.SUCCESS
                app.log.worker(self)

    def _start(
        self, app: App, done_callback: Callable[[Worker], None] | None = None
    ) -> None:
        """Start the worker.

        Args:
            app: An app instance.
            done_callback: A callback to call when the task is done.
        """
        if self._task is not None:
            return
        self.state = WorkerState.RUNNING
        self._task = asyncio.create_task(self._run(app))

        def task_done_callback(_task: asyncio.Task) -> None:
            """Run the callback.

            Called by `Task.add_done_callback`.

            Args:
                The worker's task.
            """
            if done_callback is not None:
                done_callback(self)

        self._task.add_done_callback(task_done_callback)

    def cancel(self) -> None:
        """Cancel the task."""
        self._cancelled = True
        if self._task is not None:
            self._task.cancel()
        self.cancelled_event.set()

    async def wait(self) -> ResultType:
        """Wait for the work to complete.

        Raises:
            WorkerFailed: If the Worker raised an exception.
            WorkerCancelled: If the Worker was cancelled before it completed.

        Returns:
            The return value of the work.
        """
        try:
            if active_worker.get() is self:
                raise DeadlockError(
                    "Can't call worker.wait from within the worker function!"
                )
        except LookupError:
            # Not in a worker
            pass

        if self.state == WorkerState.PENDING:
            raise WorkerError("Worker must be started before calling this method.")
        if self._task is not None:
            try:
                await self._task
            except asyncio.CancelledError as error:
                self.state = WorkerState.CANCELLED
                self._error = error
        if self.state == WorkerState.ERROR:
            assert self._error is not None
            raise WorkerFailed(self._error)
        elif self.state == WorkerState.CANCELLED:
            raise WorkerCancelled("Worker was cancelled, and did not complete.")
        return cast("ResultType", self._result)



import asyncio
import gc

import pytest

from textual.app import App, ComposeResult
from textual.containers import Vertical
from textual.widgets import Footer, Header, Label


def count_nodes() -> int:
    """Count number of references to DOMNodes."""
    dom_nodes = [
        obj
        for obj in gc.get_objects()
        if any(cls.__name__ == "DOMNode" for cls in obj.__class__.__mro__)
    ]
    print(dom_nodes)
    return len(dom_nodes)


async def run_app() -> None:
    """Run a dummy app."""

    class DummyApp(App):
        """Dummy app with a few widgets."""

        def compose(self) -> ComposeResult:
            yield Header()
            with Vertical():
                yield Label("foo")
                yield Label("bar")
            yield Footer()

    app = DummyApp()

    async with app.run_test() as pilot:
        # We should have a bunch of DOMNodes while the test is running
        assert count_nodes() > 0
        await pilot.press("ctrl+c")

    assert not app._running

    # Force a GC collection
    gc.collect()

    # After the test, all DOMNodes will have been torn down
    assert count_nodes() == 1


async def _count_app_nodes() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/4959"""

    # Should be no DOMNodes yet
    assert count_nodes() == 0

    await run_app()
    await asyncio.sleep(0)

    gc.collect()

    nodes_remaining = count_nodes()

    if nodes_remaining:
        print("NODES REMAINING")

        import objgraph

        objgraph.show_backrefs(
            [
                obj
                for obj in gc.get_objects()
                if any(cls.__name__ == "App" for cls in obj.__class__.__mro__)
            ],
            filename="graph.png",
            max_depth=15,
        )

    assert nodes_remaining == 0


# It looks like PyTest holds on to references to DOMNodes
# So this will only pass if ran in isolation
@pytest.mark.xfail
async def test_gc():
    """Regression test for https://github.com/Textualize/textual/issues/4959"""
    await _count_app_nodes()



from textual.app import App, ComposeResult
from textual.containers import Center
from textual.widgets import Button, Footer, Label


class ExampleApp(App):
    CSS = """
    Screen {
        align: center middle;
    }
    """

    BINDINGS = [("space", "toggle_button", "Toggle Button")]

    def compose(self) -> ComposeResult:
        with Center():
            yield Label("Hover the button then hit space")
        with Center():
            yield Button("Enabled", id="disable-btn")
        yield Footer()

    def action_toggle_button(self) -> None:
        self.app.bell()
        button = self.query_one("#disable-btn", Button)
        if button.disabled is False:
            button.disabled = True
            button.label = "Disabled"
        else:
            button.disabled = False
            button.label = "Enabled"


if __name__ == "__main__":
    app = ExampleApp()
    app.run()



from textual._compositor import Compositor
from textual.geometry import Region


def test_regions_to_ranges_no_regions():
    assert list(Compositor._regions_to_spans([])) == []


def test_regions_to_ranges_single_region():
    regions = [Region(0, 0, 3, 2)]
    assert list(Compositor._regions_to_spans(regions)) == [
        (0, 0, 3),
        (1, 0, 3),
    ]


def test_regions_to_ranges_partially_overlapping_regions():
    regions = [Region(0, 0, 2, 2), Region(1, 1, 2, 2)]
    assert list(Compositor._regions_to_spans(regions)) == [
        (0, 0, 2),
        (1, 0, 3),
        (2, 1, 3),
    ]


def test_regions_to_ranges_fully_overlapping_regions():
    regions = [Region(1, 1, 3, 3), Region(2, 2, 1, 1), Region(0, 2, 3, 1)]
    assert list(Compositor._regions_to_spans(regions)) == [
        (1, 1, 4),
        (2, 0, 4),
        (3, 1, 4),
    ]


def test_regions_to_ranges_disjoint_regions_different_lines():
    regions = [Region(0, 0, 2, 1), Region(2, 2, 2, 1)]
    assert list(Compositor._regions_to_spans(regions)) == [(0, 0, 2), (2, 2, 4)]


def test_regions_to_ranges_disjoint_regions_same_line():
    regions = [Region(0, 0, 1, 2), Region(2, 0, 1, 1)]
    assert list(Compositor._regions_to_spans(regions)) == [
        (0, 0, 1),
        (0, 2, 3),
        (1, 0, 1),
    ]


def test_regions_to_ranges_directly_adjacent_ranges_merged():
    regions = [Region(0, 0, 1, 2), Region(1, 0, 1, 2)]
    assert list(Compositor._regions_to_spans(regions)) == [
        (0, 0, 2),
        (1, 0, 2),
    ]



from textual.app import App, ComposeResult
from textual.widgets import Placeholder

class DisabledPlaceholderApp(App[None]):

    CSS = """
    Placeholder {
        height: 1fr;
    }
    """

    def compose(self) -> ComposeResult:
        yield Placeholder()
        yield Placeholder(disabled=True)

if __name__ == "__main__":
    DisabledPlaceholderApp().run()




import pytest

from textual.widgets.text_area import Document

TEXT = """I must not fear.
Fear is the mind-killer."""

TEXT_NEWLINE = TEXT + "\n"
TEXT_WINDOWS = TEXT.replace("\n", "\r\n")
TEXT_WINDOWS_NEWLINE = TEXT_NEWLINE.replace("\n", "\r\n")


@pytest.mark.parametrize(
    "text", [TEXT, TEXT_NEWLINE, TEXT_WINDOWS, TEXT_WINDOWS_NEWLINE]
)
def test_text(text):
    """The text we put in is the text we get out."""
    document = Document(text)
    assert document.text == text


def test_lines_newline_eof():
    document = Document(TEXT_NEWLINE)
    assert document.lines == ["I must not fear.", "Fear is the mind-killer.", ""]


def test_lines_no_newline_eof():
    document = Document(TEXT)
    assert document.lines == [
        "I must not fear.",
        "Fear is the mind-killer.",
    ]


def test_lines_windows():
    document = Document(TEXT_WINDOWS)
    assert document.lines == ["I must not fear.", "Fear is the mind-killer."]


def test_lines_windows_newline():
    document = Document(TEXT_WINDOWS_NEWLINE)
    assert document.lines == ["I must not fear.", "Fear is the mind-killer.", ""]


def test_newline_unix():
    document = Document(TEXT)
    assert document.newline == "\n"


def test_newline_windows():
    document = Document(TEXT_WINDOWS)
    assert document.newline == "\r\n"


def test_get_selected_text_no_selection():
    document = Document(TEXT)
    selection = document.get_text_range((0, 0), (0, 0))
    assert selection == ""


def test_get_selected_text_single_line():
    document = Document(TEXT_WINDOWS)
    selection = document.get_text_range((0, 2), (0, 6))
    assert selection == "must"


def test_get_selected_text_multiple_lines_unix():
    document = Document(TEXT)
    selection = document.get_text_range((0, 2), (1, 2))
    assert selection == "must not fear.\nFe"


def test_get_selected_text_multiple_lines_windows():
    document = Document(TEXT_WINDOWS)
    selection = document.get_text_range((0, 2), (1, 2))
    assert selection == "must not fear.\r\nFe"


def test_get_selected_text_including_final_newline_unix():
    document = Document(TEXT_NEWLINE)
    selection = document.get_text_range((0, 0), (2, 0))
    assert selection == TEXT_NEWLINE


def test_get_selected_text_including_final_newline_windows():
    document = Document(TEXT_WINDOWS_NEWLINE)
    selection = document.get_text_range((0, 0), (2, 0))
    assert selection == TEXT_WINDOWS_NEWLINE


def test_get_selected_text_no_newline_at_end_of_file():
    document = Document(TEXT)
    selection = document.get_text_range((0, 0), (2, 0))
    assert selection == TEXT


def test_get_selected_text_no_newline_at_end_of_file_windows():
    document = Document(TEXT_WINDOWS)
    selection = document.get_text_range((0, 0), (2, 0))
    assert selection == TEXT_WINDOWS


@pytest.mark.parametrize(
    "text", [TEXT, TEXT_NEWLINE, TEXT_WINDOWS, TEXT_WINDOWS_NEWLINE]
)
def test_index_from_location(text):
    document = Document(text)
    lines = text.split(document.newline)
    assert document.get_index_from_location((0, 0)) == 0
    assert document.get_index_from_location((0, len(lines[0]))) == len(lines[0])
    assert document.get_index_from_location((1, 0)) == len(lines[0]) + len(
        document.newline
    )
    assert document.get_index_from_location((len(lines) - 1, len(lines[-1]))) == len(
        text
    )


@pytest.mark.parametrize(
    "text", [TEXT, TEXT_NEWLINE, TEXT_WINDOWS, TEXT_WINDOWS_NEWLINE]
)
def test_location_from_index(text):
    document = Document(text)
    lines = text.split(document.newline)
    assert document.get_location_from_index(0) == (0, 0)
    assert document.get_location_from_index(len(lines[0])) == (0, len(lines[0]))
    if len(document.newline) > 1:
        assert document.get_location_from_index(len(lines[0]) + 1) == (
            0,
            len(lines[0]) + 1,
        )
    assert document.get_location_from_index(len(lines[0]) + len(document.newline)) == (
        1,
        0,
    )
    assert document.get_location_from_index(len(text)) == (
        len(lines) - 1,
        len(lines[-1]),
    )


@pytest.mark.parametrize(
    "text", [TEXT, TEXT_NEWLINE, TEXT_WINDOWS, TEXT_WINDOWS_NEWLINE]
)
def test_document_end(text):
    """The location is always what we expect."""
    document = Document(text)
    expected_line_number = (
        len(text.splitlines()) if text.endswith("\n") else len(text.splitlines()) - 1
    )
    expected_pos = 0 if text.endswith("\n") else (len(text.splitlines()[-1]))
    assert document.end == (expected_line_number, expected_pos)



from textual.app import App, ComposeResult
from textual.widgets import OptionList
from textual.widgets.option_list import Option


class LongOptionListApp(App[None]):
    def compose(self) -> ComposeResult:
        yield OptionList(*[Option(f"This is option #{n}") for n in range(100)])


if __name__ == "__main__":
    LongOptionListApp().run()
